@@ -1,5 +1,5 @@
 //
-// Copyright (C) 2011 Codership Oy <info@codership.com>
+// Copyright (C) 2011-2014 Codership Oy <info@codership.com>
 //
 
 #include "ist.hpp"
@@ -20,12 +20,13 @@ namespace
     static std::string const CONF_KEEP_KEYS     ("ist.keep_keys");
     static bool        const CONF_KEEP_KEYS_DEFAULT (true);
 
-    static std::string const CONF_SSL_KEY       ("socket.ssl_key");
-    static std::string const CONF_SSL_CERT      ("socket.ssl_cert");
-    static std::string const CONF_SSL_CA        ("socket.ssl_ca");
-    static std::string const CONF_SSL_PSWD_FILE ("socket.ssl_password_file");
+    static std::string const CONF_SSL_KEY       (COMMON_CONF_SSL_KEY);
+    static std::string const CONF_SSL_CERT      (COMMON_CONF_SSL_CERT);
+    static std::string const CONF_SSL_CA        (COMMON_CONF_SSL_CA);
+    static std::string const CONF_SSL_PSWD_FILE (COMMON_CONF_SSL_PSWD_FILE);
 
-    std::string escape_addr(const asio::ip::address& addr)
+
+    static std::string escape_addr(const asio::ip::address& addr)
     {
         if (addr.is_v4())
         {
@@ -37,7 +38,8 @@ namespace
         }
     }
 
-    static inline std::string unescape_addr(const std::string& addr)
+
+    static std::string unescape_addr(const std::string& addr)
     {
         std::string ret(addr);
         size_t pos(ret.find('['));
@@ -59,29 +61,34 @@ namespace
     }
 
 
-
     class SSLPasswordCallback
     {
     public:
         SSLPasswordCallback(const gu::Config& conf) : conf_(conf) { }
+
         std::string get_password() const
         {
             std::string   file(conf_.get(CONF_SSL_PSWD_FILE));
             std::ifstream ifs(file.c_str(), std::ios_base::in);
+
             if (ifs.good() == false)
             {
                 gu_throw_error(errno) <<
                     "could not open password file '" << file
                                                      << "'";
             }
+
             std::string ret;
             std::getline(ifs, ret);
             return ret;
         }
+
     private:
+
         const gu::Config& conf_;
     };
 
+
     static void prepare_ssl_ctx(const gu::Config& conf, asio::ssl::context& ctx)
     {
         // Here we blindly assume that ssl globals have been initialized
@@ -122,13 +129,16 @@ namespace galera
                 asmap_(asmap),
                 thread_()
             { }
+
             const gu::Config& conf()  { return conf_; }
             const std::string& peer() { return peer_; }
             wsrep_seqno_t first()     { return first_; }
             wsrep_seqno_t last()      { return last_; }
             AsyncSenderMap& asmap()   { return asmap_; }
             pthread_t       thread()  { return thread_; }
+
         private:
+
             friend class AsyncSenderMap;
             const gu::Config&  conf_;
             const std::string  peer_;
@@ -144,6 +154,17 @@ namespace galera
 std::string const
 galera::ist::Receiver::RECV_ADDR("ist.recv_addr");
 
+void
+galera::ist::register_params(gu::Config& conf)
+{
+    conf.add(Receiver::RECV_ADDR);
+    conf.add(CONF_KEEP_KEYS);
+    conf.add(CONF_SSL_KEY);
+    conf.add(CONF_SSL_CERT);
+    conf.add(CONF_SSL_CA);
+    conf.add(CONF_SSL_PSWD_FILE);
+}
+
 galera::ist::Receiver::Receiver(gu::Config& conf, const char* addr)
     :
     conf_      (conf),
@@ -169,8 +190,9 @@ galera::ist::Receiver::Receiver(gu::Config& conf, const char* addr)
         recv_addr = conf_.get(RECV_ADDR);
         return;
     }
-    catch (gu::NotFound& e) {} /* if not, check the alternative.
+    catch (gu::NotSet& e) {} /* if not, check the alternative.
                                 TODO: try to find from system. */
+
     if (addr)
     {
         try
@@ -227,7 +249,7 @@ IST_determine_recv_addr (gu::Config& conf)
             std::string ssl_key = conf.get(CONF_SSL_KEY);
             if (ssl_key.length() != 0) ssl = true;
         }
-        catch (gu::NotFound&) {}
+        catch (gu::NotSet&) {}
 
         if (ssl)
             recv_addr.insert(0, "ssl://");
@@ -302,7 +324,6 @@ galera::ist::Receiver::prepare(wsrep_seqno_t first_seqno,
         // read recv_addr_ from acceptor_ in case zero port was specified
         recv_addr_ = uri.get_scheme()
             + "://"
-            // + 
             + uri.get_host()
             + ":"
             + gu::to_string(acceptor_.local_endpoint().port());
@@ -645,7 +666,7 @@ galera::ist::Sender::~Sender()
     {
         socket_.close();
     }
-    gcache_.seqno_release();
+    gcache_.seqno_unlock();
 }
 
 void galera::ist::Sender::send(wsrep_seqno_t first, wsrep_seqno_t last)
