@@ -1,5 +1,5 @@
 //
-// Copyright (C) 2010-2012 Codership Oy <info@codership.com>
+// Copyright (C) 2010-2014 Codership Oy <info@codership.com>
 //
 
 #include "galera_common.hpp"
@@ -16,58 +16,15 @@ extern "C"
 #include <iostream>
 
 
-static inline void
-apply_wscoll(void*                    recv_ctx,
-             wsrep_apply_cb_t         apply_cb,
-             const galera::TrxHandle& trx)
-{
-    const gu::byte_t* buf(trx.write_set_buffer().first);
-    const size_t buf_len(trx.write_set_buffer().second);
-    size_t offset(0);
-    while (offset < buf_len)
-    {
-        // Skip key segment
-        std::pair<size_t, size_t> k(
-            galera::WriteSet::segment(buf, buf_len, offset));
-        offset = k.first + k.second;
-        // Data part
-        std::pair<size_t, size_t> d(
-            galera::WriteSet::segment(buf, buf_len, offset));
-        offset = d.first + d.second;
-        wsrep_status_t err = apply_cb (recv_ctx,
-                                       buf + d.first,
-                                       d.second,
-                                       trx.global_seqno());
-
-        if (gu_unlikely(err != WSREP_OK))
-        {
-            const char* const err_str(galera::wsrep_status_str(err));
-            std::ostringstream os;
-
-            os << "Failed to apply app buffer: "
-               << "seqno: "<< trx.global_seqno() << ", status: " << err_str;
-
-            galera::ApplyException ae(os.str(), err);
-
-            GU_TRACE(ae);
-
-            throw ae;
-        }
-    }
-
-    assert(offset == buf_len);
-
-    return;
-}
-
-
 static void
 apply_trx_ws(void*                    recv_ctx,
              wsrep_apply_cb_t         apply_cb,
              wsrep_commit_cb_t        commit_cb,
-             const galera::TrxHandle& trx)
+             const galera::TrxHandle& trx,
+             const wsrep_trx_meta_t&  meta)
 {
-    static const size_t max_apply_attempts(10);
+    using galera::TrxHandle;
+    static const size_t max_apply_attempts(4);
     size_t attempts(1);
 
     do
@@ -78,7 +35,9 @@ apply_trx_ws(void*                    recv_ctx,
             {
                 log_debug << "Executing TO isolated action: " << trx;
             }
-            gu_trace(apply_wscoll(recv_ctx, apply_cb, trx));
+
+            gu_trace(trx.apply(recv_ctx, apply_cb, meta));
+
             if (trx.is_toi())
             {
                 log_debug << "Done executing TO isolated action: "
@@ -95,11 +54,18 @@ apply_trx_ws(void*                    recv_ctx,
             }
             else
             {
-                wsrep_status_t err = e.wsrep_status();
+                int const err(e.status());
 
-                if (WSREP_TRX_FAIL == err)
+                if (err > 0)
                 {
-                    int const rcode(commit_cb(recv_ctx,trx.global_seqno(),false));
+                    wsrep_bool_t unused(false);
+                    int const rcode(
+                        commit_cb(
+                            recv_ctx,
+                            TrxHandle::trx_flags_to_wsrep_flags(trx.flags()),
+                            &meta,
+                            &unused,
+                            false));
                     if (WSREP_OK != rcode)
                     {
                         gu_throw_fatal << "Rollback failed. Trx: " << trx;
@@ -130,7 +96,7 @@ apply_trx_ws(void*                    recv_ctx,
         msg << "Failed to apply trx " << trx.global_seqno() << " "
             << max_apply_attempts << " times";
 
-        throw galera::ApplyException(msg.str(), WSREP_TRX_FAIL);
+        throw galera::ApplyException(msg.str(), WSREP_CB_FAILURE);
     }
 
     return;
@@ -162,12 +128,13 @@ std::ostream& galera::operator<<(std::ostream& os, ReplicatorSMM::State state)
 
 galera::ReplicatorSMM::ReplicatorSMM(const struct wsrep_init_args* args)
     :
-    logger_             (reinterpret_cast<gu_log_cb_t>(args->logger_cb)),
-    config_             (args->options),
-    set_defaults_       (config_, defaults, args->node_address),
-    trx_proto_ver_      (-1),
+    init_lib_           (reinterpret_cast<gu_log_cb_t>(args->logger_cb)),
+    config_             (),
+    init_config_        (config_, args->node_address),
+    parse_options_      (config_, args->options),
     str_proto_ver_      (-1),
     protocol_version_   (-1),
+    proto_max_          (gu::from_string<int>(config_.get(Param::proto_max))),
     state_              (S_CLOSED),
     sst_state_          (SST_NONE),
     co_mode_            (CommitOrder::from_string(
@@ -176,14 +143,18 @@ galera::ReplicatorSMM::ReplicatorSMM(const struct wsrep_init_args* args)
     state_file_         (data_dir_.length() ?
                          data_dir_+'/'+GALERA_STATE_FILE : GALERA_STATE_FILE),
     st_                 (state_file_),
+    trx_params_         (data_dir_, -1,
+                         KeySet::version(config_.get(Param::key_format))),
     uuid_               (WSREP_UUID_UNDEFINED),
     state_uuid_         (WSREP_UUID_UNDEFINED),
     state_uuid_str_     (),
     cc_seqno_           (WSREP_SEQNO_UNDEFINED),
+    pause_seqno_        (WSREP_SEQNO_UNDEFINED),
     app_ctx_            (args->app_ctx),
     view_cb_            (args->view_handler_cb),
     apply_cb_           (args->apply_cb),
     commit_cb_          (args->commit_cb),
+    unordered_cb_       (args->unordered_cb),
     sst_donate_cb_      (args->sst_donate_cb),
     synced_cb_          (args->synced_cb),
     sst_donor_          (),
@@ -194,15 +165,15 @@ galera::ReplicatorSMM::ReplicatorSMM(const struct wsrep_init_args* args)
     sst_retry_sec_      (1),
     ist_sst_            (false),
     gcache_             (config_, data_dir_),
-    gcs_                (config_, gcache_, MAX_PROTO_VER, args->proto_ver,
+    gcs_                (config_, gcache_, proto_max_, args->proto_ver,
                          args->node_name, args->node_incoming),
-    service_thd_        (gcs_),
+    service_thd_        (gcs_, gcache_),
     as_                 (0),
     gcs_as_             (gcs_, *this, gcache_),
     ist_receiver_       (config_, args->node_address),
     ist_senders_        (gcs_, gcache_),
     wsdb_               (),
-    cert_               (config_),
+    cert_               (config_, service_thd_),
     local_monitor_      (),
     apply_monitor_      (),
     commit_monitor_     (),
@@ -210,11 +181,16 @@ galera::ReplicatorSMM::ReplicatorSMM(const struct wsrep_init_args* args)
     receivers_          (),
     replicated_         (),
     replicated_bytes_   (),
+    keys_count_         (),
+    keys_bytes_         (),
+    data_bytes_         (),
+    unrd_bytes_         (),
     local_commits_      (),
     local_rollbacks_    (),
     local_cert_failures_(),
-    local_bf_aborts_    (),
     local_replays_      (),
+    causal_reads_       (),
+    preordered_id_      (),
     incoming_list_      (""),
     incoming_mutex_     (),
     wsrep_stats_        ()
@@ -258,23 +234,27 @@ galera::ReplicatorSMM::ReplicatorSMM(const struct wsrep_init_args* args)
 
     st_.get (uuid, seqno);
 
-    if (0 != args->state_uuid &&
-        *args->state_uuid != WSREP_UUID_UNDEFINED &&
-        *args->state_uuid == uuid &&
-        seqno == WSREP_SEQNO_UNDEFINED)
+    if (0 != args->state_id &&
+        args->state_id->uuid != WSREP_UUID_UNDEFINED &&
+        args->state_id->uuid == uuid                 &&
+        seqno                == WSREP_SEQNO_UNDEFINED)
     {
         /* non-trivial recovery information provided on startup, and db is safe
          * so use recovered seqno value */
-        seqno = args->state_seqno;
+        seqno = args->state_id->seqno;
     }
+
     log_debug << "End state: " << uuid << ':' << seqno << " #################";
+
     update_state_uuid (uuid);
 
     cc_seqno_ = seqno; // is it needed here?
     apply_monitor_.set_initial_position(seqno);
+
     if (co_mode_ != CommitOrder::BYPASS)
         commit_monitor_.set_initial_position(seqno);
-    cert_.assign_initial_position(seqno, trx_proto_ver_);
+
+    cert_.assign_initial_position(seqno, trx_proto_ver());
 
     build_stats_vars(wsrep_stats_);
 }
@@ -303,7 +283,8 @@ galera::ReplicatorSMM::~ReplicatorSMM()
 
 wsrep_status_t galera::ReplicatorSMM::connect(const std::string& cluster_name,
                                               const std::string& cluster_url,
-                                              const std::string& state_donor)
+                                              const std::string& state_donor,
+                                              bool  const        bootstrap)
 {
     sst_donor_ = state_donor;
     service_thd_.reset();
@@ -324,7 +305,7 @@ wsrep_status_t galera::ReplicatorSMM::connect(const std::string& cluster_name,
     gcache_.reset();
 
     if (ret == WSREP_OK &&
-        (err = gcs_.connect(cluster_name, cluster_url)) != 0)
+        (err = gcs_.connect(cluster_name, cluster_url, bootstrap)) != 0)
     {
         log_error << "gcs connect failed: " << strerror(-err);
         ret = WSREP_NODE_FAIL;
@@ -363,13 +344,15 @@ wsrep_status_t galera::ReplicatorSMM::async_recv(void* recv_ctx)
     ++receivers_;
     as_ = &gcs_as_;
 
+    bool exit_loop(false);
     wsrep_status_t retval(WSREP_OK);
 
     while (WSREP_OK == retval && state_() != S_CLOSING)
     {
         ssize_t rc;
 
-        while ((rc = as_->process(recv_ctx)) == -ECANCELED)
+        while (gu_unlikely((rc = as_->process(recv_ctx, exit_loop))
+                           == -ECANCELED))
         {
             recv_IST(recv_ctx);
             // hack: prevent fast looping until ist controlling thread
@@ -377,13 +360,27 @@ wsrep_status_t galera::ReplicatorSMM::async_recv(void* recv_ctx)
             usleep(10000);
         }
 
-        if (rc <= 0)
+        if (gu_unlikely(rc <= 0))
         {
             retval = WSREP_CONN_FAIL;
         }
+        else if (gu_unlikely(exit_loop == true))
+        {
+            assert(WSREP_OK == retval);
+
+            if (receivers_.sub_and_fetch(1) > 0)
+            {
+                log_info << "Slave thread exiting on request.";
+                break;
+            }
+
+            ++receivers_;
+            log_warn << "Refusing exit for the last slave thread.";
+        }
     }
 
-    if (receivers_.sub_and_fetch(1) == 0)
+    /* exiting loop already did proper checks */
+    if (!exit_loop && receivers_.sub_and_fetch(1) == 0)
     {
         if (state_() != S_CLOSING)
         {
@@ -396,68 +393,11 @@ wsrep_status_t galera::ReplicatorSMM::async_recv(void* recv_ctx)
         state_.shift_to(S_CLOSED);
     }
 
+    log_debug << "Slave thread exit. Return code: " << retval;
+
     return retval;
 }
 
-galera::TrxHandle*
-galera::ReplicatorSMM::local_trx(wsrep_trx_id_t trx_id)
-{
-    return wsdb_.get_trx(trx_proto_ver_, uuid_, trx_id, false);
-}
-
-galera::TrxHandle*
-galera::ReplicatorSMM::local_trx(wsrep_trx_handle_t* handle, bool create)
-{
-    TrxHandle* trx;
-    assert(handle != 0);
-
-    if (handle->opaque != 0)
-    {
-        trx = reinterpret_cast<TrxHandle*>(handle->opaque);
-        assert(trx->trx_id() == handle->trx_id ||
-               wsrep_trx_id_t(-1) == handle->trx_id);
-        trx->ref();
-    }
-    else
-    {
-        trx = wsdb_.get_trx(trx_proto_ver_, uuid_, handle->trx_id, create);
-        handle->opaque = trx;
-    }
-
-    return trx;
-}
-
-
-void galera::ReplicatorSMM::unref_local_trx(TrxHandle* trx)
-{
-    wsdb_.unref_trx(trx);
-}
-
-
-void galera::ReplicatorSMM::discard_local_trx(wsrep_trx_id_t trx_id)
-{
-    wsdb_.discard_trx(trx_id);
-}
-
-
-galera::TrxHandle*
-galera::ReplicatorSMM::local_conn_trx(wsrep_conn_id_t conn_id, bool create)
-{
-    return wsdb_.get_conn_query(trx_proto_ver_, uuid_, conn_id, create);
-}
-
-
-void galera::ReplicatorSMM::discard_local_conn_trx(wsrep_conn_id_t conn_id)
-{
-    wsdb_.discard_conn_query(conn_id);
-}
-
-
-void galera::ReplicatorSMM::discard_local_conn(wsrep_conn_id_t conn_id)
-{
-    wsdb_.discard_conn(conn_id);
-}
-
 
 void galera::ReplicatorSMM::apply_trx(void* recv_ctx, TrxHandle* trx)
 {
@@ -472,38 +412,58 @@ void galera::ReplicatorSMM::apply_trx(void* recv_ctx, TrxHandle* trx)
 
     gu_trace(apply_monitor_.enter(ao));
     trx->set_state(TrxHandle::S_APPLYING);
-    gu_trace(apply_trx_ws(recv_ctx, apply_cb_, commit_cb_, *trx));
-    // at this point any exception in apply_trx_ws() is fatal, not
-    // catching anything.
+
+    wsrep_trx_meta_t meta = {{state_uuid_, trx->global_seqno() },
+                             trx->depends_seqno()};
+
+    gu_trace(apply_trx_ws(recv_ctx, apply_cb_, commit_cb_, *trx, meta));
+    /* at this point any exception in apply_trx_ws() is fatal, not
+     * catching anything. */
+
     if (gu_likely(co_mode_ != CommitOrder::BYPASS))
     {
         gu_trace(commit_monitor_.enter(co));
-        trx->set_state(TrxHandle::S_COMMITTING);
-        if (gu_unlikely (WSREP_OK != commit_cb_(recv_ctx, trx->global_seqno(),
-                                                true)))
-            gu_throw_fatal << "Commit failed. Trx: " << trx;
+    }
+    trx->set_state(TrxHandle::S_COMMITTING);
 
-        commit_monitor_.leave(co);
-        trx->set_state(TrxHandle::S_COMMITTED);
-    }
-    else
+    wsrep_bool_t exit_loop(false);
+    wsrep_cb_status_t const rcode(
+        commit_cb_(
+            recv_ctx,
+            TrxHandle::trx_flags_to_wsrep_flags(trx->flags()),
+            &meta,
+            &exit_loop,
+            true));
+
+    if (gu_unlikely (rcode > 0))
+        gu_throw_fatal << "Commit failed. Trx: " << trx;
+
+    if (gu_likely(co_mode_ != CommitOrder::BYPASS))
     {
-        trx->set_state(TrxHandle::S_COMMITTING);
-        if (gu_unlikely (WSREP_OK != commit_cb_(recv_ctx, trx->global_seqno(),
-                                                true)))
-            gu_throw_fatal << "Commit failed. Trx: " << trx;
-        trx->set_state(TrxHandle::S_COMMITTED);
+        commit_monitor_.leave(co);
     }
-    apply_monitor_.leave(ao);
+    trx->set_state(TrxHandle::S_COMMITTED);
 
     if (trx->local_seqno() != -1)
     {
         // trx with local seqno -1 originates from IST (or other source not gcs)
         report_last_committed(cert_.set_trx_committed(trx));
     }
+
+    /* For now need to keep it inside apply monitor to ensure all processing
+     * ends by the time monitors are drained because of potential gcache
+     * cleanup (and loss of the writeset buffer). Perhaps unordered monitor
+     * is needed here. */
+    trx->unordered(recv_ctx, unordered_cb_);
+
+    apply_monitor_.leave(ao);
+
+    trx->set_exit_loop(exit_loop);
 }
 
-wsrep_status_t galera::ReplicatorSMM::replicate(TrxHandle* trx)
+
+wsrep_status_t galera::ReplicatorSMM::replicate(TrxHandle* trx,
+                                                wsrep_trx_meta_t* meta)
 {
     if (state_() < S_JOINED) return WSREP_TRX_FAIL;
 
@@ -521,24 +481,43 @@ wsrep_status_t galera::ReplicatorSMM::replicate(TrxHandle* trx)
         return retval;
     }
 
-    trx->set_last_seen_seqno(last_committed());
-    trx->flush(0);
-    trx->set_state(TrxHandle::S_REPLICATING);
-
-    const MappedBuffer& wscoll(trx->write_set_collection());
-    ssize_t rcode;
+    WriteSetNG::GatherVector actv;
 
     gcs_action act;
-    act.size = wscoll.size();
     act.type = GCS_ACT_TORDERED;
 #ifndef NDEBUG
     act.seqno_g = GCS_SEQNO_ILL;
 #endif
 
+    if (trx->new_version())
+    {
+        act.buf  = NULL;
+        act.size = trx->write_set_out().gather(trx->source_id(),
+                                               trx->conn_id(),
+                                               trx->trx_id(),
+                                               actv);
+    }
+    else
+    {
+        trx->set_last_seen_seqno(last_committed());
+        assert (trx->last_seen_seqno() >= 0);
+        trx->flush(0);
+
+        const MappedBuffer& wscoll(trx->write_set_collection());
+
+        act.buf  = &wscoll[0];
+        act.size = wscoll.size();
+
+        assert (act.buf != NULL);
+        assert (act.size > 0);
+    }
+
+    trx->set_state(TrxHandle::S_REPLICATING);
+
+    ssize_t rcode(-1);
+
     do
     {
-        act.buf = &wscoll[0];
-        assert(act.buf);
         assert(act.seqno_g == GCS_SEQNO_ILL);
 
         const ssize_t gcs_handle(gcs_.schedule());
@@ -551,15 +530,30 @@ wsrep_status_t galera::ReplicatorSMM::replicate(TrxHandle* trx)
         }
 
         trx->set_gcs_handle(gcs_handle);
-        trx->unlock();
 
-        rcode = gcs_.repl(act, true);
+        if (trx->new_version())
+        {
+            trx->set_last_seen_seqno(last_committed());
+            assert(trx->last_seen_seqno() >= 0);
+            trx->unlock();
+            assert (act.buf == NULL); // just a sanity check
+            rcode = gcs_.replv(actv, act, true);
+        }
+        else
+        {
+            assert(trx->last_seen_seqno() >= 0);
+            trx->unlock();
+            assert (act.buf != NULL);
+            rcode = gcs_.repl(act, true);
+        }
 
         trx->lock();
     }
     while (rcode == -EAGAIN && trx->state() != TrxHandle::S_MUST_ABORT &&
            (usleep(1000), true));
 
+    assert(trx->last_seen_seqno() >= 0);
+
     if (rcode < 0)
     {
         if (rcode != -EINTR)
@@ -570,6 +564,7 @@ wsrep_status_t galera::ReplicatorSMM::replicate(TrxHandle* trx)
 
         assert(rcode != -EINTR || trx->state() == TrxHandle::S_MUST_ABORT);
         assert(act.seqno_l == GCS_SEQNO_ILL && act.seqno_g == GCS_SEQNO_ILL);
+        assert(NULL == act.buf || !trx->new_version());
 
         if (trx->state() != TrxHandle::S_MUST_ABORT)
         {
@@ -580,12 +575,22 @@ wsrep_status_t galera::ReplicatorSMM::replicate(TrxHandle* trx)
         goto must_abort;
     }
 
-    assert(act.buf);
-    assert(act.seqno_l != GCS_SEQNO_ILL && act.seqno_g != GCS_SEQNO_ILL);
+    assert(act.buf != NULL);
+    assert(act.size == rcode);
+    assert(act.seqno_l != GCS_SEQNO_ILL);
+    assert(act.seqno_g != GCS_SEQNO_ILL);
 
     ++replicated_;
-    replicated_bytes_ += wscoll.size();
+    replicated_bytes_ += rcode;
     trx->set_gcs_handle(-1);
+
+    if (trx->new_version())
+    {
+        gu_trace(trx->unserialize(static_cast<const gu::byte_t*>(act.buf),
+                                  act.size, 0));
+        trx->update_stats(keys_count_, keys_bytes_, data_bytes_, unrd_bytes_);
+    }
+
     trx->set_received(act.buf, act.seqno_l, act.seqno_g);
 
     if (trx->state() == TrxHandle::S_MUST_ABORT)
@@ -601,6 +606,12 @@ wsrep_status_t galera::ReplicatorSMM::replicate(TrxHandle* trx)
             apply_monitor_.self_cancel(ao);
             if (co_mode_ !=CommitOrder::BYPASS) commit_monitor_.self_cancel(co);
         }
+        else if (meta != 0)
+        {
+            meta->gtid.uuid  = state_uuid_;
+            meta->gtid.seqno = trx->global_seqno();
+            meta->depends_on = trx->depends_seqno();
+        }
 
         if (trx->state() == TrxHandle::S_MUST_ABORT) goto must_abort;
     }
@@ -609,6 +620,8 @@ wsrep_status_t galera::ReplicatorSMM::replicate(TrxHandle* trx)
         retval = WSREP_OK;
     }
 
+    assert(trx->last_seen_seqno() >= 0);
+
     return retval;
 }
 
@@ -620,7 +633,6 @@ galera::ReplicatorSMM::abort_trx(TrxHandle* trx)
 
     log_debug << "aborting trx " << *trx << " " << trx;
 
-    ++local_bf_aborts_;
 
     switch (trx->state())
     {
@@ -682,15 +694,24 @@ galera::ReplicatorSMM::abort_trx(TrxHandle* trx)
 }
 
 
-wsrep_status_t galera::ReplicatorSMM::pre_commit(TrxHandle* trx)
+wsrep_status_t galera::ReplicatorSMM::pre_commit(TrxHandle*        trx,
+                                                 wsrep_trx_meta_t* meta)
 {
+    assert(trx->state() == TrxHandle::S_REPLICATING);
+    assert(trx->local_seqno()  > -1);
+    assert(trx->global_seqno() > -1);
+    assert(trx->last_seen_seqno() >= 0);
+
+    if (meta != 0)
+    {
+        meta->gtid.uuid  = state_uuid_;
+        meta->gtid.seqno = trx->global_seqno();
+        meta->depends_on = trx->depends_seqno();
+    }
     // State should not be checked here: If trx has been replicated,
     // it has to be certified and potentially applied. #528
     // if (state_() < S_JOINED) return WSREP_TRX_FAIL;
 
-    assert(trx->state() == TrxHandle::S_REPLICATING);
-    assert(trx->local_seqno() > -1 && trx->global_seqno() > -1);
-
     wsrep_status_t retval(cert(trx));
 
     if (gu_unlikely(retval != WSREP_OK))
@@ -821,10 +842,21 @@ wsrep_status_t galera::ReplicatorSMM::replay_trx(TrxHandle* trx, void* trx_ctx)
 
         try
         {
-            gu_trace(apply_trx_ws(trx_ctx, apply_cb_, commit_cb_, *trx));
+            wsrep_trx_meta_t meta = {{state_uuid_, trx->global_seqno() },
+                                     trx->depends_seqno()};
 
-            if (gu_unlikely(WSREP_OK != commit_cb_(trx_ctx, trx->global_seqno(),
-                                                   true)))
+            gu_trace(apply_trx_ws(trx_ctx, apply_cb_, commit_cb_, *trx, meta));
+
+            wsrep_bool_t unused(false);
+            wsrep_cb_status_t rcode(
+                commit_cb_(
+                    trx_ctx,
+                    TrxHandle::trx_flags_to_wsrep_flags(trx->flags()),
+                    &meta,
+                    &unused,
+                    true));
+
+            if (gu_unlikely(rcode > 0))
                 gu_throw_fatal << "Commit failed. Trx: " << trx;
         }
         catch (gu::Exception& e)
@@ -868,9 +900,9 @@ wsrep_status_t galera::ReplicatorSMM::post_commit(TrxHandle* trx)
     if (co_mode_ != CommitOrder::BYPASS) commit_monitor_.leave(co);
 
     ApplyOrder ao(*trx);
+    report_last_committed(cert_.set_trx_committed(trx));
     apply_monitor_.leave(ao);
 
-    report_last_committed(cert_.set_trx_committed(trx));
     trx->set_state(TrxHandle::S_COMMITTED);
 
     ++local_commits_;
@@ -900,7 +932,7 @@ wsrep_status_t galera::ReplicatorSMM::post_rollback(TrxHandle* trx)
 }
 
 
-wsrep_status_t galera::ReplicatorSMM::causal_read(wsrep_seqno_t* seqno)
+wsrep_status_t galera::ReplicatorSMM::causal_read(wsrep_gtid_t* gtid)
 {
     wsrep_seqno_t cseq(static_cast<wsrep_seqno_t>(gcs_.caused()));
 
@@ -930,7 +962,11 @@ wsrep_status_t galera::ReplicatorSMM::causal_read(wsrep_seqno_t* seqno)
         {
             apply_monitor_.wait(cseq, wait_until);
         }
-        if (seqno != 0) *seqno = cseq;
+        if (gtid != 0)
+        {
+            gtid->uuid = state_uuid_;
+            gtid->seqno = cseq;
+        }
         ++causal_reads_;
         return WSREP_OK;
     }
@@ -942,8 +978,16 @@ wsrep_status_t galera::ReplicatorSMM::causal_read(wsrep_seqno_t* seqno)
 }
 
 
-wsrep_status_t galera::ReplicatorSMM::to_isolation_begin(TrxHandle* trx)
+wsrep_status_t galera::ReplicatorSMM::to_isolation_begin(TrxHandle*        trx,
+                                                         wsrep_trx_meta_t* meta)
 {
+    if (meta != 0)
+    {
+        meta->gtid.uuid  = state_uuid_;
+        meta->gtid.seqno = trx->global_seqno();
+        meta->depends_on = trx->depends_seqno();
+    }
+
     assert(trx->state() == TrxHandle::S_REPLICATING);
     assert(trx->trx_id() == static_cast<wsrep_trx_id_t>(-1));
     assert(trx->local_seqno() > -1 && trx->global_seqno() > -1);
@@ -977,8 +1021,7 @@ wsrep_status_t galera::ReplicatorSMM::to_isolation_begin(TrxHandle* trx)
     case WSREP_TRX_FAIL:
         // Apply monitor is released in cert() in case of failure.
         trx->set_state(TrxHandle::S_ABORTING);
-        // Called now from cert()
-        // report_last_committed();
+        // Called now from cert(): report_last_committed();
         break;
     default:
         log_error << "unrecognized retval "
@@ -1002,25 +1045,131 @@ wsrep_status_t galera::ReplicatorSMM::to_isolation_end(TrxHandle* trx)
     CommitOrder co(*trx, co_mode_);
     if (co_mode_ != CommitOrder::BYPASS) commit_monitor_.leave(co);
     ApplyOrder ao(*trx);
+    report_last_committed(cert_.set_trx_committed(trx));
     apply_monitor_.leave(ao);
 
     st_.mark_safe();
-    report_last_committed(cert_.set_trx_committed(trx));
+
+    return WSREP_OK;
+}
+
+namespace galera
+{
+
+static WriteSetOut*
+writeset_from_handle (wsrep_po_handle_t& handle,
+                      const TrxHandle::Params& trx_params)
+{
+    WriteSetOut* ret = reinterpret_cast<WriteSetOut*>(handle.opaque);
+
+    if (NULL == ret)
+    {
+        try
+        {
+            ret = new WriteSetOut(
+//                gu::String<256>(trx_params.working_dir_) << '/' << &handle,
+                trx_params.working_dir_, wsrep_trx_id_t(&handle),
+                /* key format is not essential since we're not adding keys */
+                KeySet::version(trx_params.key_format_), NULL, 0);
+
+            handle.opaque = ret;
+        }
+        catch (std::bad_alloc& ba)
+        {
+            gu_throw_error(ENOMEM) << "Could not create WriteSetOut";
+        }
+    }
+
+    return ret;
+}
+
+} /* namespace galera */
+
+wsrep_status_t
+galera::ReplicatorSMM::preordered_collect(wsrep_po_handle_t&            handle,
+                                          const struct wsrep_buf* const data,
+                                          size_t                  const count,
+                                          bool                    const copy)
+{
+    if (gu_unlikely(trx_params_.version_ < WS_NG_VERSION))
+        return WSREP_NOT_IMPLEMENTED;
+
+    WriteSetOut* const ws(writeset_from_handle(handle, trx_params_));
+
+    for (size_t i(0); i < count; ++i)
+    {
+        ws->append_data(data[i].ptr, data[i].len, copy);
+    }
 
     return WSREP_OK;
 }
 
 
 wsrep_status_t
-galera::ReplicatorSMM::sst_sent(const wsrep_uuid_t& uuid, wsrep_seqno_t seqno)
+galera::ReplicatorSMM::preordered_commit(wsrep_po_handle_t&            handle,
+                                         const wsrep_uuid_t&           source,
+                                         uint64_t                const flags,
+                                         int                     const pa_range,
+                                         bool                    const commit)
 {
+    if (gu_unlikely(trx_params_.version_ < WS_NG_VERSION))
+        return WSREP_NOT_IMPLEMENTED;
+
+    WriteSetOut* const ws(writeset_from_handle(handle, trx_params_));
+
+    if (gu_likely(true == commit))
+    {
+        ws->set_flags (WriteSetNG::wsrep_flags_to_ws_flags(flags));
+
+        /* by loooking at trx_id we should be able to detect gaps / lost events
+         * (however resending is not implemented yet). Something like
+         *
+         * wsrep_trx_id_t const trx_id(cert_.append_preordered(source, ws));
+         *
+         * begs to be here. */
+        wsrep_trx_id_t const trx_id(preordered_id_.add_and_fetch(1));
+
+        WriteSetNG::GatherVector actv;
+
+        size_t const actv_size(ws->gather(source, 0, trx_id, actv));
+
+        ws->set_preordered (pa_range); // also adds CRC
+
+        int rcode;
+        do
+        {
+            rcode = gcs_.sendv(actv, actv_size, GCS_ACT_TORDERED, false);
+        }
+        while (rcode == -EAGAIN && (usleep(1000), true));
+
+        if (rcode < 0)
+            gu_throw_error(-rcode)
+                << "Replication of preordered writeset failed.";
+    }
+
+    delete ws;
+    handle.opaque = NULL;
+
+    return WSREP_OK;
+}
+
+
+wsrep_status_t
+galera::ReplicatorSMM::sst_sent(const wsrep_gtid_t& state_id, int const rcode)
+{
+    assert (rcode <= 0);
+    assert (rcode == 0 || state_id.seqno == WSREP_SEQNO_UNDEFINED);
+    assert (rcode != 0 || state_id.seqno >= 0);
+
     if (state_() != S_DONOR)
     {
         log_error << "sst sent called when not SST donor, state " << state_();
         return WSREP_CONN_FAIL;
     }
 
-    if (uuid != state_uuid_ && seqno >= 0)
+    gcs_seqno_t seqno(rcode ? rcode : state_id.seqno);
+
+    if (state_id.uuid != state_uuid_ && seqno >= 0)
     {
         // state we have sent no longer corresponds to the current group state
         // mark an error
@@ -1029,7 +1178,7 @@ galera::ReplicatorSMM::sst_sent(const wsrep_uuid_t& uuid, wsrep_seqno_t seqno)
 
     try {
         // #557 - remove this if() when we return back to joining after SST
-        if (!ist_sst_ || seqno < 0) gcs_.join(seqno);
+        if (!ist_sst_ || rcode < 0) gcs_.join(seqno);
         ist_sst_ = false;
         return WSREP_OK;
     }
@@ -1092,7 +1241,11 @@ void galera::ReplicatorSMM::process_commit_cut(wsrep_seqno_t seq,
     LocalOrder lo(seqno_l);
 
     gu_trace(local_monitor_.enter(lo));
-    cert_.purge_trxs_upto(seq);
+
+    if (seq >= cc_seqno_) /* Refs #782. workaround for
+                           * assert(seqno >= seqno_released_) in gcache. */
+        cert_.purge_trxs_upto(seq, true);
+
     local_monitor_.leave(lo);
     log_debug << "Got commit cut from GCS: " << seq;
 }
@@ -1102,16 +1255,20 @@ void galera::ReplicatorSMM::establish_protocol_versions (int proto_ver)
     switch (proto_ver)
     {
     case 1:
-        trx_proto_ver_ = 1;
+        trx_params_.version_ = 1;
         str_proto_ver_ = 0;
         break;
     case 2:
-        trx_proto_ver_ = 1;
+        trx_params_.version_ = 1;
         str_proto_ver_ = 1;
         break;
     case 3:
     case 4:
-        trx_proto_ver_ = 2;
+        trx_params_.version_ = 2;
+        str_proto_ver_ = 1;
+        break;
+    case 5:
+        trx_params_.version_ = 3;
         str_proto_ver_ = 1;
         break;
     default:
@@ -1121,8 +1278,8 @@ void galera::ReplicatorSMM::establish_protocol_versions (int proto_ver)
     };
 
     protocol_version_ = proto_ver;
-    log_debug << "REPL Protocols: " << protocol_version_ << " ("
-              << trx_proto_ver_ << ", " << str_proto_ver_ << ")";
+    log_info << "REPL Protocols: " << protocol_version_ << " ("
+              << trx_params_.version_ << ", " << str_proto_ver_ << ")";
 }
 
 static bool
@@ -1191,8 +1348,8 @@ galera::ReplicatorSMM::process_conf_change(void*                    recv_ctx,
     }
 
     bool const          st_required(state_transfer_required(view_info));
-    wsrep_seqno_t const group_seqno(view_info.seqno);
-    const wsrep_uuid_t& group_uuid (view_info.uuid);
+    wsrep_seqno_t const group_seqno(view_info.state_id.seqno);
+    const wsrep_uuid_t& group_uuid (view_info.state_id.uuid);
 
     if (st_required)
     {
@@ -1204,18 +1361,19 @@ galera::ReplicatorSMM::process_conf_change(void*                    recv_ctx,
         if (S_CONNECTED != state_()) state_.shift_to(S_CONNECTED);
     }
 
-    void*   app_req(0);
-    ssize_t app_req_len(0);
+    void*  app_req(0);
+    size_t app_req_len(0);
 
     const_cast<wsrep_view_info_t&>(view_info).state_gap = st_required;
-    view_cb_(app_ctx_, recv_ctx, &view_info, 0, 0, &app_req, &app_req_len);
+    wsrep_cb_status_t const rcode(
+        view_cb_(app_ctx_, recv_ctx, &view_info, 0, 0, &app_req, &app_req_len));
 
-    if (app_req_len < 0)
+    if (WSREP_CB_SUCCESS != rcode)
     {
+        assert(app_req_len <= 0);
         close();
-        gu_throw_fatal << "View callback failed: " << -app_req_len << " ("
-                       << strerror(-app_req_len) << "). This is unrecoverable, "
-                       << "restart required.";
+        gu_throw_fatal << "View callback failed. This is unrecoverable, "
+            "restart required.";
     }
     else if (st_required && 0 == app_req_len && state_uuid_ != group_uuid)
     {
@@ -1231,7 +1389,14 @@ galera::ReplicatorSMM::process_conf_change(void*                    recv_ctx,
 
         // we have to reset cert initial position here, SST does not contain
         // cert index yet (see #197).
-        cert_.assign_initial_position(group_seqno, trx_proto_ver_);
+        cert_.assign_initial_position(group_seqno, trx_params_.version_);
+        // at this point there is no ongoing master or slave transactions
+        // and no new requests to service thread should be possible
+
+        service_thd_.flush();             // make sure service thd is idle
+
+        if (upto > 0) gcache_.seqno_release(upto); // make sure all gcache
+                                                   // buffers are released
 
         // record state seqno, needed for IST on DONOR
         cc_seqno_ = group_seqno;
@@ -1240,6 +1405,7 @@ galera::ReplicatorSMM::process_conf_change(void*                    recv_ctx,
 
         if (st_required && app_wants_st)
         {
+            // GCache::Seqno_reset() happens here
             request_state_transfer (recv_ctx,
                                     group_uuid, group_seqno, app_req,
                                     app_req_len);
@@ -1371,18 +1537,19 @@ void galera::ReplicatorSMM::process_sync(wsrep_seqno_t seqno_l)
 
 wsrep_seqno_t galera::ReplicatorSMM::pause()
 {
-    try
-    {
-        gu_trace(local_monitor_.lock());
-    }
-    catch (gu::Exception& e)
-    {
-        if (e.get_errno() == EALREADY) return cert_.position();
-        throw;
-    }
+    // Grab local seqno for local_monitor_
+    wsrep_seqno_t const local_seqno(
+        static_cast<wsrep_seqno_t>(gcs_.local_sequence()));
+    LocalOrder lo(local_seqno);
+    local_monitor_.enter(lo);
 
+    // Local monitor should take care that concurrent
+    // pause requests are enqueued
+    assert(pause_seqno_ == WSREP_SEQNO_UNDEFINED);
+    pause_seqno_ = local_seqno;
+
+    // Get drain seqno from cert index
     wsrep_seqno_t const ret(cert_.position());
-
     apply_monitor_.drain(ret);
     assert (apply_monitor_.last_left() >= ret);
 
@@ -1394,30 +1561,25 @@ wsrep_seqno_t galera::ReplicatorSMM::pause()
 
     st_.set(state_uuid_, ret);
 
-    log_info << "Provider paused at " << state_uuid_ << ':' << ret;
+    log_info << "Provider paused at " << state_uuid_ << ':' << ret
+             << " (" << pause_seqno_ << ")";
 
     return ret;
 }
 
 void galera::ReplicatorSMM::resume()
 {
+    assert(pause_seqno_ != WSREP_SEQNO_UNDEFINED);
+    if (pause_seqno_ == WSREP_SEQNO_UNDEFINED)
+    {
+        gu_throw_error(EALREADY) << "tried to resume unpaused provider";
+    }
+
     st_.set(state_uuid_, WSREP_SEQNO_UNDEFINED);
-
-    try
-    {
-        gu_trace(local_monitor_.unlock());
-    }
-    catch (gu::Exception& e)
-    {
-        if (e.get_errno() == EBUSY)
-        {
-            /* monitor is still locked, restore saved state */
-            st_.set(state_uuid_, cert_.position());
-            return;
-        }
-        throw;
-    }
-
+    log_info << "resuming provider at " << pause_seqno_;
+    LocalOrder lo(pause_seqno_);
+    pause_seqno_ = WSREP_SEQNO_UNDEFINED;
+    local_monitor_.leave(lo);
     log_info << "Provider resumed.";
 }
 
@@ -1476,9 +1638,10 @@ wsrep_status_t galera::ReplicatorSMM::cert(TrxHandle* trx)
     assert(trx->state() == TrxHandle::S_REPLICATING ||
            trx->state() == TrxHandle::S_MUST_CERT_AND_REPLAY);
 
-    assert(trx->local_seqno()     != WSREP_SEQNO_UNDEFINED &&
-           trx->global_seqno()    != WSREP_SEQNO_UNDEFINED &&
-           trx->last_seen_seqno() != WSREP_SEQNO_UNDEFINED);
+    assert(trx->local_seqno()     != WSREP_SEQNO_UNDEFINED);
+    assert(trx->global_seqno()    != WSREP_SEQNO_UNDEFINED);
+    assert(trx->last_seen_seqno() >= 0);
+    assert(trx->last_seen_seqno() < trx->global_seqno());
 
     trx->set_state(TrxHandle::S_CERTIFYING);
 
@@ -1532,9 +1695,10 @@ wsrep_status_t galera::ReplicatorSMM::cert(TrxHandle* trx)
             {
                 if (gu_unlikely(trx->is_toi())) // small sanity check
                 {
-                    log_error << "Certification failed for TO isolated action: "
-                              << *trx;
-                    assert(0); // should never happen
+                    // may happen on configuration change
+                    log_warn << "Certification failed for TO isolated action: "
+                             << *trx;
+                    assert(0);
                 }
                 apply_monitor_.self_cancel(ao);
                 if (co_mode_ != CommitOrder::BYPASS)
@@ -1552,8 +1716,7 @@ wsrep_status_t galera::ReplicatorSMM::cert(TrxHandle* trx)
         /*! @todo: benchmark both variants on SMP */
         gcache_.seqno_assign (trx->action(),
                               trx->global_seqno(),
-                              trx->depends_seqno(),
-                              trx->is_local());  // frees local actions
+                              trx->depends_seqno());
 
         local_monitor_.leave(lo);
     }
@@ -1591,10 +1754,8 @@ wsrep_status_t galera::ReplicatorSMM::cert_for_aborted(TrxHandle* trx)
         }
         retval = WSREP_TRX_FAIL;
 
-        gcache_.seqno_assign (trx->action(),
-                              trx->global_seqno(),
-                              -1,
-                              trx->is_local());
+        gcache_.seqno_assign (trx->action(), trx->global_seqno(), -1);
+
         break;
     }
     return retval;
