@@ -1,7 +1,9 @@
 //
-// Copyright (C) 2010 Codership Oy <info@codership.com>
+// Copyright (C) 2010-2014 Codership Oy <info@codership.com>
 //
 
+#include "key_data.hpp"
+
 #if defined(GALERA_MULTIMASTER)
 #include "replicator_smm.hpp"
 #define REPL_CLASS galera::ReplicatorSMM
@@ -13,7 +15,8 @@
 
 #include <cassert>
 
-using galera::Key;
+
+using galera::KeyOS;
 using galera::WriteSet;
 using galera::TrxHandle;
 using galera::TrxHandleLock;
@@ -39,6 +42,10 @@ wsrep_status_t galera_init(wsrep_t* gh, const struct wsrep_init_args* args)
     {
         log_error << e.what();
     }
+    catch (gu::NotFound& e)
+    {
+        /* Unrecognized parameter (logged by gu::Config::set()) */
+    }
 #ifdef NDEBUG
     catch (...)
     {
@@ -53,13 +60,28 @@ wsrep_status_t galera_init(wsrep_t* gh, const struct wsrep_init_args* args)
 extern "C"
 uint64_t galera_capabilities(wsrep_t* gh)
 {
-    return (WSREP_CAP_MULTI_MASTER      |
-            WSREP_CAP_CERTIFICATION     |
-            WSREP_CAP_PARALLEL_APPLYING |
-            WSREP_CAP_TRX_REPLAY        |
-            WSREP_CAP_ISOLATION         |
-            WSREP_CAP_PAUSE             |
-            WSREP_CAP_CAUSAL_READS);
+    assert(gh != 0);
+    assert(gh->ctx != 0);
+
+    static uint64_t const v4_caps(WSREP_CAP_MULTI_MASTER         |
+                                  WSREP_CAP_CERTIFICATION        |
+                                  WSREP_CAP_PARALLEL_APPLYING    |
+                                  WSREP_CAP_TRX_REPLAY           |
+                                  WSREP_CAP_ISOLATION            |
+                                  WSREP_CAP_PAUSE                |
+                                  WSREP_CAP_CAUSAL_READS);
+
+    static uint64_t const v5_caps(WSREP_CAP_INCREMENTAL_WRITESET |
+                                  WSREP_CAP_UNORDERED            |
+                                  WSREP_CAP_PREORDERED);
+
+    uint64_t caps(v4_caps);
+
+    REPL_CLASS * repl(reinterpret_cast< REPL_CLASS * >(gh->ctx));
+
+    if (repl->repl_proto_ver() >= 5) caps |= v5_caps;
+
+    return caps;
 }
 
 
@@ -82,11 +104,12 @@ void galera_tear_down(wsrep_t* gh)
 extern "C"
 wsrep_status_t galera_parameters_set (wsrep_t* gh, const char* params)
 {
-    assert(gh != 0);
+    assert(gh != 0); // cppcheck-suppress nullPointer
     assert(gh->ctx != 0);
 
     REPL_CLASS * repl(reinterpret_cast< REPL_CLASS * >(gh->ctx));
 
+    // cppcheck-suppress nullPointer
     if (gh)
     {
         try
@@ -94,6 +117,11 @@ wsrep_status_t galera_parameters_set (wsrep_t* gh, const char* params)
             wsrep_set_params (*repl, params);
             return WSREP_OK;
         }
+        catch (gu::NotFound&)
+        {
+            log_warn << "Unrecognized parameter in '" << params << "'";
+            return WSREP_WARNING;
+        }
         catch (std::exception& e)
         {
             log_debug << e.what(); // better logged in wsrep_set_params
@@ -133,10 +161,11 @@ char* galera_parameters_get (wsrep_t* gh)
 
 
 extern "C"
-wsrep_status_t galera_connect (wsrep_t*    gh,
-                               const char* cluster_name,
-                               const char* cluster_url,
-                               const char* state_donor)
+wsrep_status_t galera_connect (wsrep_t*     gh,
+                               const char*  cluster_name,
+                               const char*  cluster_url,
+                               const char*  state_donor,
+                               wsrep_bool_t bootstrap)
 {
     assert(gh != 0);
     assert(gh->ctx != 0);
@@ -146,7 +175,7 @@ wsrep_status_t galera_connect (wsrep_t*    gh,
     try
     {
         return repl->connect(cluster_name, cluster_url,
-                             state_donor ? state_donor : "");
+                             state_donor ? state_donor : "", bootstrap);
     }
     catch (std::exception& e)
     {
@@ -228,6 +257,66 @@ wsrep_status_t galera_recv(wsrep_t *gh, void *recv_ctx)
     return WSREP_FATAL;
 }
 
+static TrxHandle*
+get_local_trx(REPL_CLASS* const        repl,
+              wsrep_ws_handle_t* const handle,
+              bool const               create)
+{
+    TrxHandle* trx;
+    assert(handle != 0);
+
+    if (handle->opaque != 0)
+    {
+        trx = static_cast<TrxHandle*>(handle->opaque);
+        assert(trx->trx_id() == handle->trx_id ||
+               wsrep_trx_id_t(-1) == handle->trx_id);
+        trx->ref();
+    }
+    else
+    {
+        trx = repl->get_local_trx(handle->trx_id, create);
+        handle->opaque = trx;
+    }
+
+    return trx;
+}
+
+extern "C"
+wsrep_status_t galera_replay_trx(wsrep_t*            gh,
+                                 wsrep_ws_handle_t*  trx_handle,
+                                 void*               recv_ctx)
+{
+    assert(gh != 0);
+    assert(gh->ctx != 0);
+
+    REPL_CLASS * repl(reinterpret_cast< REPL_CLASS * >(gh->ctx));
+    TrxHandle* trx(get_local_trx(repl, trx_handle, false));
+    assert(trx != 0);
+
+    wsrep_status_t retval;
+
+    try
+    {
+        TrxHandleLock lock(*trx);
+        retval = repl->replay_trx(trx, recv_ctx);
+    }
+    catch (std::exception& e)
+    {
+        log_warn << "failed to replay trx: " << *trx;
+        log_warn << e.what();
+        retval = WSREP_CONN_FAIL;
+    }
+    catch (...)
+    {
+        log_fatal << "non-standard exception";
+        retval = WSREP_FATAL;
+    }
+
+    repl->unref_local_trx(trx);
+
+    return retval;
+}
+
 
 extern "C"
 wsrep_status_t galera_abort_pre_commit(wsrep_t*       gh,
@@ -239,7 +328,7 @@ wsrep_status_t galera_abort_pre_commit(wsrep_t*       gh,
 
     REPL_CLASS *   repl(reinterpret_cast< REPL_CLASS * >(gh->ctx));
     wsrep_status_t retval;
-    TrxHandle*     trx(repl->local_trx(victim_trx));
+    TrxHandle*     trx(repl->get_local_trx(victim_trx));
 
     if (!trx) return WSREP_OK;
 
@@ -265,20 +354,29 @@ wsrep_status_t galera_abort_pre_commit(wsrep_t*       gh,
     return retval;
 }
 
+static inline void
+discard_local_trx(REPL_CLASS*        repl,
+                  wsrep_ws_handle_t* ws_handle,
+                  TrxHandle*         trx)
+{
+    repl->unref_local_trx(trx);
+    repl->discard_local_trx(trx);
+    ws_handle->opaque = 0;
+}
 
 extern "C"
 wsrep_status_t galera_post_commit (wsrep_t*            gh,
-                                   wsrep_trx_handle_t* trx_handle)
+                                   wsrep_ws_handle_t*  ws_handle)
 {
     assert(gh != 0);
     assert(gh->ctx != 0);
 
     REPL_CLASS * repl(reinterpret_cast< REPL_CLASS * >(gh->ctx));
-    TrxHandle* trx(repl->local_trx(trx_handle, false));
+    TrxHandle* trx(get_local_trx(repl, ws_handle, false));
 
     if (trx == 0)
     {
-        log_debug << "trx " << trx_handle->trx_id << " not found";
+        log_debug << "trx " << ws_handle->trx_id << " not found";
         return WSREP_OK;
     }
 
@@ -300,9 +398,7 @@ wsrep_status_t galera_post_commit (wsrep_t*            gh,
         retval = WSREP_FATAL;
     }
 
-    repl->unref_local_trx(trx);
-    repl->discard_local_trx(trx->trx_id());
-    trx_handle->opaque = 0;
+    discard_local_trx(repl, ws_handle, trx);
 
     return retval;
 }
@@ -310,17 +406,17 @@ wsrep_status_t galera_post_commit (wsrep_t*            gh,
 
 extern "C"
 wsrep_status_t galera_post_rollback(wsrep_t*            gh,
-                                    wsrep_trx_handle_t* trx_handle)
+                                    wsrep_ws_handle_t*  ws_handle)
 {
     assert(gh != 0);
     assert(gh->ctx != 0);
 
     REPL_CLASS * repl(reinterpret_cast< REPL_CLASS * >(gh->ctx));
-    TrxHandle* trx(repl->local_trx(trx_handle, false));
+    TrxHandle* trx(get_local_trx(repl, ws_handle, false));
 
     if (trx == 0)
     {
-        log_debug << "trx " << trx_handle->trx_id << " not found";
+        log_debug << "trx " << ws_handle->trx_id << " not found";
         return WSREP_OK;
     }
 
@@ -342,31 +438,45 @@ wsrep_status_t galera_post_rollback(wsrep_t*            gh,
         retval = WSREP_FATAL;
     }
 
-    repl->unref_local_trx(trx);
-    repl->discard_local_trx(trx->trx_id());
-    trx_handle->opaque = 0;
+    discard_local_trx(repl, ws_handle, trx);
 
     return retval;
 }
 
 
+static inline void
+append_data_array (TrxHandle*              const trx,
+                   const struct wsrep_buf* const data,
+                   size_t                  const count,
+                   wsrep_data_type_t       const type,
+                   bool                    const copy)
+{
+    for (size_t i(0); i < count; ++i)
+    {
+        trx->append_data(data[i].ptr, data[i].len, type, copy);
+    }
+}
+
+
 extern "C"
-wsrep_status_t galera_pre_commit(wsrep_t*            gh,
-                                 wsrep_conn_id_t     conn_id,
-                                 wsrep_trx_handle_t* trx_handle,
-                                 const void*         rbr_data,
-                                 size_t              rbr_data_len,
-                                 uint64_t            flags __attribute__((unused)) ,
-                                 wsrep_seqno_t*      global_seqno)
+wsrep_status_t galera_pre_commit(wsrep_t*           const gh,
+                                 wsrep_conn_id_t    const conn_id,
+                                 wsrep_ws_handle_t* const trx_handle,
+                                 uint32_t           const flags,
+                                 wsrep_trx_meta_t*  const meta)
 {
     assert(gh != 0);
     assert(gh->ctx != 0);
 
-    *global_seqno = WSREP_SEQNO_UNDEFINED;
+    if (meta != 0)
+    {
+        meta->gtid = WSREP_GTID_UNDEFINED;
+        meta->depends_on = WSREP_SEQNO_UNDEFINED;
+    }
 
     REPL_CLASS * repl(reinterpret_cast< REPL_CLASS * >(gh->ctx));
 
-    TrxHandle* trx(repl->local_trx(trx_handle, rbr_data != 0));
+    TrxHandle* trx(get_local_trx(repl, trx_handle, /*rbr_data != 0*/ false));
 
     if (trx == 0)
     {
@@ -380,21 +490,19 @@ wsrep_status_t galera_pre_commit(wsrep_t*            gh,
     {
         TrxHandleLock lock(*trx);
         trx->set_conn_id(conn_id);
-        trx->append_data(rbr_data, rbr_data_len);
-        trx->set_flags(
-            TrxHandle::F_COMMIT |
-            ((flags & WSREP_FLAG_PA_SAFE) ? 0 : TrxHandle::F_PA_UNSAFE)
-            );
+//        /* rbr_data should clearly persist over pre_commit() call */
+//        append_data_array (trx, rbr_data, rbr_data_len, false, false);
+        trx->set_flags(TrxHandle::wsrep_flags_to_trx_flags(flags));
 
-        retval = repl->replicate(trx);
+        retval = repl->replicate(trx, meta);
 
         assert((!(retval == WSREP_OK || retval == WSREP_BF_ABORT) ||
                 trx->global_seqno() > 0));
 
         if (retval == WSREP_OK)
         {
-            *global_seqno = trx->global_seqno();
-            retval = repl->pre_commit(trx);
+            assert(trx->last_seen_seqno() >= 0);
+            retval = repl->pre_commit(trx, meta);
         }
 
         assert(retval == WSREP_OK || retval == WSREP_TRX_FAIL ||
@@ -418,29 +526,18 @@ wsrep_status_t galera_pre_commit(wsrep_t*            gh,
 
 
 extern "C"
-wsrep_status_t galera_append_query(wsrep_t*            gh,
-                                   wsrep_trx_handle_t* trx_handle,
-                                   const char*         query,
-                                   const time_t        timeval,
-                                   const uint32_t      randseed)
-{
-    log_warn << "galera_append_query() is deprecated";
-    return WSREP_CONN_FAIL;
-}
-
-
-extern "C"
-wsrep_status_t galera_append_key(wsrep_t*            gh,
-                                 wsrep_trx_handle_t* trx_handle,
-                                 const wsrep_key_t*  key,
-                                 size_t              key_len,
-                                 bool                shared)
+wsrep_status_t galera_append_key(wsrep_t*           const gh,
+                                 wsrep_ws_handle_t* const trx_handle,
+                                 const wsrep_key_t* const keys,
+                                 size_t             const keys_num,
+                                 wsrep_key_type_t   const key_type,
+                                 wsrep_bool_t       const copy)
 {
     assert(gh != 0);
     assert(gh->ctx != 0);
 
     REPL_CLASS * repl(reinterpret_cast< REPL_CLASS * >(gh->ctx));
-    TrxHandle* trx(repl->local_trx(trx_handle, true));
+    TrxHandle* trx(get_local_trx(repl, trx_handle, true));
     assert(trx != 0);
 
     wsrep_status_t retval;
@@ -448,14 +545,13 @@ wsrep_status_t galera_append_key(wsrep_t*            gh,
     try
     {
         TrxHandleLock lock(*trx);
-        for (size_t i(0); i < key_len; ++i)
+        for (size_t i(0); i < keys_num; ++i)
         {
-            galera::Key k(repl->trx_proto_ver(),
-                          key[i].key_parts,
-                          key[i].key_parts_len,
-                          // std::min(key[i].key_parts_len, size_t(2)),
-                          (shared == true ? galera::Key::F_SHARED : 0));
-            // log_info << k;
+            galera::KeyData k (repl->trx_proto_ver(),
+                               keys[i].key_parts,
+                               keys[i].key_parts_num,
+                               key_type,
+                               copy);
             trx->append_key(k);
         }
         retval = WSREP_OK;
@@ -475,20 +571,58 @@ wsrep_status_t galera_append_key(wsrep_t*            gh,
     return retval;
 }
 
-
 extern "C"
-wsrep_status_t galera_append_data(wsrep_t*            wsrep,
-                                  wsrep_trx_handle_t* trx_handle,
-                                  const void*         data,
-                                  size_t              data_len)
+wsrep_status_t galera_append_data(wsrep_t*                const wsrep,
+                                  wsrep_ws_handle_t*      const trx_handle,
+                                  const struct wsrep_buf* const data,
+                                  size_t                  const count,
+                                  wsrep_data_type_t       const type,
+                                  wsrep_bool_t            const copy)
 {
-    return WSREP_NOT_IMPLEMENTED;
+    assert(wsrep != 0);
+    assert(wsrep->ctx != 0);
+    assert(data != NULL);
+    assert(count > 0);
+
+    if (data == NULL)
+    {
+        // no data to replicate
+        return WSREP_OK;
+    }
+
+    REPL_CLASS * repl(reinterpret_cast< REPL_CLASS * >(wsrep->ctx));
+    TrxHandle* trx(get_local_trx(repl, trx_handle, true));
+    assert(trx != 0);
+
+    wsrep_status_t retval;
+
+    try
+    {
+        TrxHandleLock lock(*trx);
+        if (WSREP_DATA_ORDERED == type)
+            append_data_array(trx, data, count, type, copy);
+        retval = WSREP_OK;
+    }
+    catch (std::exception& e)
+    {
+        log_warn << e.what();
+        retval = WSREP_CONN_FAIL;
+    }
+    catch (...)
+    {
+        log_fatal << "non-standard exception";
+        retval = WSREP_FATAL;
+    }
+
+    repl->unref_local_trx(trx);
+
+    return retval;
 }
 
 
 extern "C"
-wsrep_status_t galera_causal_read(wsrep_t*       wsrep,
-                                  wsrep_seqno_t* seqno)
+wsrep_status_t galera_causal_read(wsrep_t*      const wsrep,
+                                  wsrep_gtid_t* const gtid)
 {
     assert(wsrep != 0);
     assert(wsrep->ctx != 0);
@@ -497,7 +631,7 @@ wsrep_status_t galera_causal_read(wsrep_t*       wsrep,
     wsrep_status_t retval;
     try
     {
-        retval = repl->causal_read(seqno);
+        retval = repl->causal_read(gtid);
     }
     catch (std::exception& e)
     {
@@ -514,8 +648,8 @@ wsrep_status_t galera_causal_read(wsrep_t*       wsrep,
 
 
 extern "C"
-wsrep_status_t galera_free_connection(wsrep_t*              gh,
-                                      const wsrep_conn_id_t conn_id)
+wsrep_status_t galera_free_connection(wsrep_t*        const gh,
+                                      wsrep_conn_id_t const conn_id)
 {
     assert(gh != 0);
     assert(gh->ctx != 0);
@@ -541,13 +675,13 @@ wsrep_status_t galera_free_connection(wsrep_t*              gh,
 
 
 extern "C"
-wsrep_status_t galera_to_execute_start(wsrep_t*           gh,
-                                       wsrep_conn_id_t    conn_id,
-                                       const wsrep_key_t* key,
-                                       size_t             key_len,
-                                       const void*        query,
-                                       size_t             query_len,
-                                       wsrep_seqno_t*     global_seqno)
+wsrep_status_t galera_to_execute_start(wsrep_t*                const gh,
+                                       wsrep_conn_id_t         const conn_id,
+                                       const wsrep_key_t*      const keys,
+                                       size_t                  const keys_num,
+                                       const struct wsrep_buf* const data,
+                                       size_t                  const count,
+                                       wsrep_trx_meta_t*       const meta)
 {
     assert(gh != 0);
     assert(gh->ctx != 0);
@@ -562,25 +696,28 @@ wsrep_status_t galera_to_execute_start(wsrep_t*           gh,
     try
     {
         TrxHandleLock lock(*trx);
-        for (size_t i(0); i < key_len; ++i)
+        for (size_t i(0); i < keys_num; ++i)
         {
-            trx->append_key(Key(repl->trx_proto_ver(),
-                                key[i].key_parts,
-                                key[i].key_parts_len, 0));
+            galera::KeyData k(repl->trx_proto_ver(),
+                              keys[i].key_parts,
+                              keys[i].key_parts_num, WSREP_KEY_EXCLUSIVE,false);
+            trx->append_key(k);
         }
-        trx->append_data(query, query_len);
-        trx->set_flags(TrxHandle::F_COMMIT | TrxHandle::F_ISOLATION);
 
-        retval = repl->replicate(trx);
+        append_data_array(trx, data, count, WSREP_DATA_ORDERED, false);
+
+        trx->set_flags(TrxHandle::wsrep_flags_to_trx_flags(
+                           WSREP_FLAG_COMMIT |
+                           WSREP_FLAG_ISOLATION));
+
+        retval = repl->replicate(trx, meta);
 
         assert((retval == WSREP_OK && trx->global_seqno() > 0) ||
                (retval != WSREP_OK && trx->global_seqno() < 0));
 
-        *global_seqno = trx->global_seqno();
-
         if (retval == WSREP_OK)
         {
-            retval = repl->to_isolation_begin(trx);
+            retval = repl->to_isolation_begin(trx, meta);
         }
     }
     catch (std::exception& e)
@@ -598,7 +735,7 @@ wsrep_status_t galera_to_execute_start(wsrep_t*           gh,
     {
         repl->discard_local_conn_trx(conn_id); // trx is not needed anymore
 
-        if (*global_seqno < 0) // no seqno -> no index -> no automatic purging
+        if (trx->global_seqno() < 0) // no seqno -> no index -> no automatic purging
         {
             trx->unref(); // implicit destructor
         }
@@ -609,7 +746,8 @@ wsrep_status_t galera_to_execute_start(wsrep_t*           gh,
 
 
 extern "C"
-wsrep_status_t galera_to_execute_end(wsrep_t* gh, wsrep_conn_id_t conn_id)
+wsrep_status_t galera_to_execute_end(wsrep_t*        const gh,
+                                     wsrep_conn_id_t const conn_id)
 {
     assert(gh != 0);
     assert(gh->ctx != 0);
@@ -642,66 +780,105 @@ wsrep_status_t galera_to_execute_end(wsrep_t* gh, wsrep_conn_id_t conn_id)
 }
 
 
-extern "C"
-wsrep_status_t galera_replay_trx(wsrep_t*            gh,
-                                 wsrep_trx_handle_t* trx_handle,
-                                 void*               recv_ctx)
+extern "C" wsrep_status_t
+galera_preordered_collect (wsrep_t* const gh,
+                           wsrep_po_handle_t*      const handle,
+                           const struct wsrep_buf* const data,
+                           size_t                  const count,
+                           wsrep_bool_t            const copy)
 {
     assert(gh != 0);
     assert(gh->ctx != 0);
+    assert(handle != 0);
+    assert(data != 0);
+    assert(count > 0);
 
     REPL_CLASS * repl(reinterpret_cast< REPL_CLASS * >(gh->ctx));
-    TrxHandle* trx(repl->local_trx(trx_handle, false));
-    assert(trx != 0);
-
-    wsrep_status_t retval;
 
     try
     {
-        TrxHandleLock lock(*trx);
-        retval = repl->replay_trx(trx, recv_ctx);
+        return repl->preordered_collect(*handle, data, count, copy);
     }
     catch (std::exception& e)
     {
-        log_warn << "failed to replay trx: " << *trx;
         log_warn << e.what();
-        retval = WSREP_CONN_FAIL;
+        return WSREP_TRX_FAIL;
     }
     catch (...)
     {
         log_fatal << "non-standard exception";
-        retval = WSREP_FATAL;
+        return WSREP_FATAL;
     }
+}
 
-    repl->unref_local_trx(trx);
 
-    return retval;
+extern "C" wsrep_status_t
+galera_preordered_commit (wsrep_t* const gh,
+                          wsrep_po_handle_t*      const handle,
+                          const wsrep_uuid_t*     const source_id,
+                          uint32_t                const flags,
+                          int                     const pa_range,
+                          wsrep_bool_t            const commit)
+{
+    assert(gh != 0);
+    assert(gh->ctx != 0);
+    assert(handle != 0);
+    assert(source_id != 0 || false == commit);
+    assert(pa_range  >= 0 || false == commit);
+
+    REPL_CLASS * repl(reinterpret_cast< REPL_CLASS * >(gh->ctx));
+
+    try
+    {
+        return repl->preordered_commit(*handle, *source_id, flags, pa_range,
+                                       commit);
+    }
+    catch (std::exception& e)
+    {
+        log_warn << e.what();
+        return WSREP_TRX_FAIL;
+    }
+    catch (...)
+    {
+        log_fatal << "non-standard exception";
+        return WSREP_FATAL;
+    }
 }
 
 
 extern "C"
-wsrep_status_t galera_sst_sent (wsrep_t*            gh,
-                                const wsrep_uuid_t* uuid,
-                                wsrep_seqno_t       seqno)
+wsrep_status_t galera_sst_sent (wsrep_t*            const gh,
+                                const wsrep_gtid_t* const state_id,
+                                int                 const rcode)
 {
-    assert(gh != 0);
-    assert(gh->ctx != 0);
+    assert(gh       != 0);
+    assert(gh->ctx  != 0);
+    assert(state_id != 0);
+    assert(rcode    <= 0);
+
     REPL_CLASS * repl(reinterpret_cast< REPL_CLASS * >(gh->ctx));
-    return repl->sst_sent(*uuid, seqno);
+
+    return repl->sst_sent(*state_id, rcode);
 }
 
 
 extern "C"
-wsrep_status_t galera_sst_received (wsrep_t*            gh,
-                                    const wsrep_uuid_t* uuid,
-                                    wsrep_seqno_t       seqno,
-                                    const char*         state,
-                                    size_t              state_len)
+wsrep_status_t galera_sst_received (wsrep_t*            const gh,
+                                    const wsrep_gtid_t* const state_id,
+                                    const void*         const state,
+                                    size_t              const state_len,
+                                    int                 const rcode)
 {
-    assert(gh != 0);
-    assert(gh->ctx != 0);
+    assert(gh       != 0);
+    assert(gh->ctx  != 0);
+    assert(state_id != 0);
+    assert(rcode    <= 0);
+
     REPL_CLASS * repl(reinterpret_cast< REPL_CLASS * >(gh->ctx));
-    return repl->sst_received(*uuid, seqno, state, state_len);
+
+    if (rcode < 0) { assert(state_id->seqno == WSREP_SEQNO_UNDEFINED); }
+
+    return repl->sst_received(*state_id, state, state_len, rcode);
 }
 
 
@@ -720,7 +897,9 @@ struct wsrep_stats_var* galera_stats_get (wsrep_t* gh)
 {
     assert(gh != 0);
     assert(gh->ctx != 0);
-    REPL_CLASS * repl(reinterpret_cast< REPL_CLASS * >(gh->ctx));
+
+    REPL_CLASS* repl(reinterpret_cast< REPL_CLASS * >(gh->ctx));
+
     return const_cast<struct wsrep_stats_var*>(repl->stats_get());
 }
 
@@ -733,6 +912,18 @@ void galera_stats_free (wsrep_t* gh, struct wsrep_stats_var* s)
 }
 
 
+extern "C"
+void galera_stats_reset (wsrep_t* gh)
+{
+    assert(gh != 0);
+    assert(gh->ctx != 0);
+
+    REPL_CLASS* repl(reinterpret_cast< REPL_CLASS * >(gh->ctx));
+
+    repl->stats_reset();
+}
+
+
 extern "C"
 wsrep_seqno_t galera_pause (wsrep_t* gh)
 {
@@ -817,10 +1008,11 @@ wsrep_status_t galera_resync (wsrep_t* gh)
 
 
 extern "C"
-wsrep_status_t galera_lock (wsrep_t* gh,
-                            const char* name,
-                            int64_t     owner,
-                            int64_t     timeout)
+wsrep_status_t galera_lock (wsrep_t*     gh,
+                            const char*  name,
+                            wsrep_bool_t shared,
+                            uint64_t     owner,
+                            int64_t      timeout)
 {
     assert(gh != 0);
     assert(gh->ctx != 0);
@@ -829,9 +1021,9 @@ wsrep_status_t galera_lock (wsrep_t* gh,
 
 
 extern "C"
-wsrep_status_t galera_unlock (wsrep_t* gh,
+wsrep_status_t galera_unlock (wsrep_t*    gh,
                               const char* name,
-                              int64_t     owner)
+                              uint64_t    owner)
 {
     assert(gh != 0);
     assert(gh->ctx != 0);
@@ -840,9 +1032,9 @@ wsrep_status_t galera_unlock (wsrep_t* gh,
 
 
 extern "C"
-bool galera_is_locked (wsrep_t* gh,
+bool galera_is_locked (wsrep_t*      gh,
                        const char*   name,
-                       int64_t*      owner,
+                       uint64_t*     owner,
                        wsrep_uuid_t* node)
 {
     assert(gh != 0);
@@ -865,18 +1057,20 @@ static wsrep_t galera_str = {
     &galera_post_rollback,
     &galera_replay_trx,
     &galera_abort_pre_commit,
-    &galera_append_query,
     &galera_append_key,
     &galera_append_data,
     &galera_causal_read,
     &galera_free_connection,
     &galera_to_execute_start,
     &galera_to_execute_end,
+    &galera_preordered_collect,
+    &galera_preordered_commit,
     &galera_sst_sent,
     &galera_sst_received,
     &galera_snapshot,
     &galera_stats_get,
     &galera_stats_free,
+    &galera_stats_reset,
     &galera_pause,
     &galera_resume,
     &galera_desync,
@@ -885,7 +1079,7 @@ static wsrep_t galera_str = {
     &galera_unlock,
     &galera_is_locked,
     "Galera",
-    GALERA_VER"(r"GALERA_REV")",
+    GALERA_VER "(r" GALERA_REV ")",
     "Codership Oy <info@codership.com>",
     &galera_tear_down,
     NULL,
