@@ -3,7 +3,7 @@
 /*!
  * @file GaleraUtils main header file
  *
- * $Id: galerautils.hpp 3196 2013-08-09 07:26:28Z vlad $
+ * $Id: galerautils.hpp 3354 2013-11-02 18:13:24Z teemu $
  */
 
 #ifndef _galerautils_hpp_
@@ -20,7 +20,7 @@
 #include "gu_macros.hpp"
 #include "gu_utils.hpp"
 #include "gu_convert.hpp"
-#include "gu_string.hpp"
+//#include "gu_string.hpp"
 #include "gu_config.hpp"
 #include "gu_prodcons.hpp"
 
