@@ -3,7 +3,7 @@
 /**
  * @file Endian conversion templates for serialization
  *
- * $Id: gu_byteswap.hpp 3196 2013-08-09 07:26:28Z vlad $
+ * $Id: gu_byteswap.hpp 3354 2013-11-02 18:13:24Z teemu $
  */
 
 #ifndef _gu_byteswap_hpp_
@@ -35,16 +35,29 @@ template <> GU_FORCE_INLINE uint16_t gtoh (const uint16_t& val)
     return  gtoh16(val);
 }
 
-template <> GU_FORCE_INLINE uint32_t gtoh (const uint32_t& val)
+template <> GU_FORCE_INLINE unsigned int gtoh (const unsigned int& val)
 {
     return  gtoh32(val);
 }
 
-template <> GU_FORCE_INLINE uint64_t gtoh (const uint64_t& val)
+#if __LONG_MAX__ == __INT_MAX__
+template <> GU_FORCE_INLINE unsigned long gtoh (const unsigned long& val)
+{
+    return  gtoh32(val);
+}
+#elif __LONG_MAX__ == __LONG_LONG_MAX__
+template <> GU_FORCE_INLINE unsigned long gtoh (const unsigned long& val)
 {
     return  gtoh64(val);
 }
+#else
+# error can not determine size of long
+#endif
 
+template <> GU_FORCE_INLINE unsigned long long gtoh (const unsigned long long& val)
+{
+    return  gtoh64(val);
+}
 
 template <typename T> T htog (const T& val) { return gtoh<T>(val); }
 
