@@ -3,7 +3,7 @@
 /**
  * @file Miscellaneous C++-related macros
  *
- * $Id: gu_macros.hpp 3196 2013-08-09 07:26:28Z vlad $
+ * $Id: gu_macros.hpp 3434 2013-12-23 15:22:18Z alex $
  */
 
 #ifndef _gu_macros_hpp_
@@ -19,4 +19,12 @@ extern "C" { static void (* const GU_SIG_IGN)(int) = SIG_IGN; }
 extern "C" { static const void* const GU_MAP_FAILED = MAP_FAILED; }
 #endif
 
+namespace gu
+{
+    template<bool> struct CompileAssert {};
+} /* namespace gu */
+
+#define GU_COMPILE_ASSERT(expr,msg) \
+    typedef gu::CompileAssert<(bool(expr))> msg[bool(expr) ? 1 : -1] __attribute__((unused))
+
 #endif /* _gu_macros_hpp_ */
