@@ -8,7 +8,7 @@
  *
  * gu_rand_seed_int()  goes for srand() and rand_r()
  *
- * $Id: gu_rand.c 3107 2013-06-11 09:03:53Z teemu $
+ * $Id: gu_rand.c 3307 2013-09-21 11:57:52Z alex $
  */
 
 #include "gu_rand.h"
@@ -21,12 +21,7 @@ struct gu_rse
     long long   time;
     const void* heap_ptr;
     const void* stack_ptr;
-#ifdef WITH_VALGRIND
-    /* To suppress uninitialized memory valgrind error */
     long        pid;
-#else
-    pid_t       pid;
-#endif /* WITH_VALGRIND */
 };
 
 typedef struct gu_rse gu_rse_t;
