@@ -1,6 +1,6 @@
-// Copyright (C) 2009 Codership Oy <info@codership.com>
+// Copyright (C) 2009-2014 Codership Oy <info@codership.com>
 
-// $Id: gu_tests++.hpp 3196 2013-08-09 07:26:28Z vlad $
+// $Id: gu_tests++.hpp 3455 2014-02-01 18:11:09Z alex $
 
 /*!
  * @file: package specific part of the main test file.
@@ -10,21 +10,35 @@
 
 #define LOG_FILE "gu_tests++.log"
 
+#include "gu_vector_test.hpp"
 #include "gu_string_test.hpp"
+#include "gu_vlq_test.hpp"
+#include "gu_digest_test.hpp"
+#include "gu_mem_pool_test.hpp"
+#include "gu_alloc_test.hpp"
+#include "gu_rset_test.hpp"
+#include "gu_string_utils_test.hpp"
 #include "gu_uri_test.hpp"
+#include "gu_config_test.hpp"
 #include "gu_net_test.hpp"
 #include "gu_datetime_test.hpp"
-#include "gu_vlq_test.hpp"
 
 typedef Suite *(*suite_creator_t)(void);
 
 static suite_creator_t suites[] =
 {
+    gu_vector_suite,
     gu_string_suite,
+    gu_vlq_suite,
+    gu_digest_suite,
+    gu_mem_pool_suite,
+    gu_alloc_suite,
+    gu_rset_suite,
+    gu_string_utils_suite,
     gu_uri_suite,
+    gu_config_suite,
     gu_net_suite,
     gu_datetime_suite,
-    gu_vlq_suite,
     0
 };
 
