@@ -1,5 +1,5 @@
 /*
- * Copyright (C) 2009 Codership Oy <info@codership.com>
+ * Copyright (C) 2009-2014 Codership Oy <info@codership.com>
  */
 
 /*!
@@ -100,6 +100,14 @@ namespace gcomm
          */
         static std::string const SocketSslCompression;
 
+        /*!
+         * @brief Algorithm for message checksums:
+         * 0 - none    (backward compatible)
+         * 1 - CRC-32  (backward compatible)
+         * 2 - CRC-32C (optimized and potentially HW-accelerated on Intel CPUs)
+         */
+        static std::string const SocketChecksum;
+
         /*!
          * @brief GMCast scheme for transport URI ("gmcast")
          */
@@ -187,6 +195,12 @@ namespace gcomm
          */
         static std::string const GMCastIsolate;
 
+        /*!
+         * @brief Segment identifier for segmentation.
+         */
+        static std::string const GMCastSegment;
+
+
         /*!
          * @brief EVS scheme for transport URI ("evs")
          */
@@ -406,12 +420,12 @@ namespace gcomm
          */
         static std::string const PcWeight;
 
+        static void register_params(gu::Config&);
     };
 
 
     // Helper templates to read configuration parameters.
 
-
     template <typename T> T _conf_param(const gu::URI& uri,
                                         const std::string& param,
                                         const T* default_value = 0,
@@ -425,12 +439,14 @@ namespace gcomm
         }
         catch (gu::NotFound& e)
         {
+            // cppcheck-suppress nullPointer
             if (default_value == 0)
             {
                 gu_throw_error(EINVAL)
                     << "param " << param << " not found from uri "
                     << uri.to_string();
             }
+            // cppcheck-suppress nullPointer
             ret = *default_value;
         }
 
@@ -486,7 +502,8 @@ namespace gcomm
                                                const T& max_value)
 
     {
-        return _conf_param(uri, param, &default_value, reinterpret_cast<const T*>(0), &max_value);
+        return _conf_param(uri, param, &default_value,
+                           reinterpret_cast<const T*>(0), &max_value);
     }
 
     template <typename T> T conf_param_def_range(const gu::URI& uri,
