@@ -52,6 +52,9 @@ gcomm::GMCast::GMCast(Protonet& net, const gu::URI& uri)
     version_(check_range(Conf::GMCastVersion,
                          param<int>(conf_, uri, Conf::GMCastVersion, "0"),
                          0, max_version_ + 1)),
+    segment_ (check_range(Conf::GMCastSegment,
+                          param<int>(conf_, uri, Conf::GMCastSegment, "0"),
+                          0, 255)),
     my_uuid_      (0, 0),
     use_ssl_      (param<bool>(conf_, uri, Conf::SocketUseSsl, "false")),
     // @todo: technically group name should be in path component
@@ -75,8 +78,9 @@ gcomm::GMCast::GMCast(Protonet& net, const gu::URI& uri)
     relaying_     (false),
     isolate_      (false),
     proto_map_    (new ProtoMap()),
-    mcast_tree_   (),
     relay_set_    (),
+    segment_map_  (),
+    self_index_   (std::numeric_limits<size_t>::max()),
     time_wait_    (param<gu::datetime::Period>(conf_, uri, Conf::GMCastTimeWait, "PT5S")),
     check_period_ ("PT0.5S"),
     peer_timeout_ (param<gu::datetime::Period>(conf_, uri, Conf::GMCastPeerTimeout, "PT3S")),
@@ -141,7 +145,7 @@ gcomm::GMCast::GMCast(Protonet& net, const gu::URI& uri)
         {
             port = conf_.get(BASE_PORT_KEY);
         }
-        catch (gu::NotFound&)
+        catch (gu::NotSet&)
         {
             // if no base port configured, try port from the connection address
             try { port = uri_.get_port(); } catch (gu::NotSet&) {}
@@ -150,24 +154,21 @@ gcomm::GMCast::GMCast(Protonet& net, const gu::URI& uri)
         listen_addr_ += ":" + port;
     }
 
-    // if (!conf_.has(BASE_PORT_KEY)) {
-        conf_.set(BASE_PORT_KEY, port);
-    // }
+    conf_.set(BASE_PORT_KEY, port);
 
-        listen_addr_ = gu::net::resolve(listen_addr_).to_string();
+    listen_addr_ = gu::net::resolve(listen_addr_).to_string();
     // resolving sets scheme to tcp, have to rewrite for ssl
     if (use_ssl_ == true)
     {
         listen_addr_.replace(0, 3, gcomm::SSL_SCHEME);
     }
 
-    if (initial_addrs_.find(listen_addr_) != initial_addrs_.end())
+    std::set<std::string>::iterator iaself(initial_addrs_.find(listen_addr_));
+    if (iaself != initial_addrs_.end())
     {
-        gu_throw_error(EINVAL) << "connect address points to listen address '"
-                               << listen_addr_
-                               << "', check that cluster address '"
-                               << uri.get_host() << ":" << port
-                               << "' is correct";
+        log_debug << "removing own listen address '" << *iaself
+                  << "' from initial address list";
+        initial_addrs_.erase(iaself);
     }
 
     if (mcast_addr_ != "")
@@ -192,7 +193,7 @@ gcomm::GMCast::GMCast(Protonet& net, const gu::URI& uri)
     conf_.set(Conf::GMCastTimeWait, gu::to_string(time_wait_));
     conf_.set(Conf::GMCastMCastTTL, gu::to_string(mcast_ttl_));
     conf_.set(Conf::GMCastPeerTimeout, gu::to_string(peer_timeout_));
-
+    conf_.set(Conf::GMCastSegment, gu::to_string<int>(segment_));
 }
 
 gcomm::GMCast::~GMCast()
@@ -326,7 +327,7 @@ void gcomm::GMCast::close(bool force)
     delete listener_;
     listener_ = 0;
 
-    mcast_tree_.clear();
+    segment_map_.clear();
     for (ProtoMap::iterator
              i = proto_map_->begin(); i != proto_map_->end(); ++i)
     {
@@ -364,7 +365,9 @@ void gcomm::GMCast::gmcast_accept()
         version_, tp,
         listener_->listen_addr() /* listen_addr */,
         "", mcast_addr_,
-        uuid(), group_name_);
+        uuid(),
+        segment_,
+        group_name_);
     std::pair<ProtoMap::iterator, bool> ret =
         proto_map_->insert(std::make_pair(tp->id(), peer));
 
@@ -418,6 +421,7 @@ void gcomm::GMCast::gmcast_connect(const std::string& remote_addr)
         remote_addr,
         mcast_addr_,
         uuid(),
+        segment_,
         group_name_);
 
     std::pair<ProtoMap::iterator, bool> ret =
@@ -513,9 +517,12 @@ void gcomm::GMCast::handle_established(Proto* est)
         AddrList::iterator i(pending_addrs_.find(est->remote_addr()));
         if (i != pending_addrs_.end())
         {
-            log_warn << self_string()
-                     << " address '" << est->remote_addr()
-                     << "' points to own listening address, blacklisting";
+            if (addr_blacklist_.find(est->remote_addr()) == addr_blacklist_.end())
+            {
+                log_warn << self_string()
+                         << " address '" << est->remote_addr()
+                         << "' points to own listening address, blacklisting";
+            }
             pending_addrs_.erase(i);
             addr_blacklist_.insert(make_pair(est->remote_addr(),
                                              AddrEntry(gu::datetime::Date::now(),
@@ -835,30 +842,47 @@ void gcomm::GMCast::update_addresses()
 
     // Build multicast tree
     log_debug << self_string() << " --- mcast tree begin ---";
-    mcast_tree_.clear();
+    segment_map_.clear();
+
+    Segment& local_segment(segment_map_[segment_]);
 
     if (mcast_ != 0)
     {
         log_debug << mcast_addr_;
-        mcast_tree_.push_back(mcast_.get());
+        local_segment.push_back(mcast_.get());
     }
 
+    self_index_ = 0;
     for (ProtoMap::const_iterator i(proto_map_->begin()); i != proto_map_->end();
          ++i)
     {
         const Proto& p(*ProtoMap::value(i));
 
-        log_debug << "Proto: " << p.state() << " " << p.remote_addr()
-                  << " " << p.mcast_addr();
+        log_debug << "Proto: " << p;
 
-        if (p.state() == Proto::S_OK &&
-            (p.mcast_addr() == "" ||
-             p.mcast_addr() != mcast_addr_))
+        if (p.remote_segment() == segment_)
         {
-            log_debug << p.remote_addr();
-            mcast_tree_.push_back(p.socket().get());
+            if (p.state() == Proto::S_OK &&
+                (p.mcast_addr() == "" ||
+                 p.mcast_addr() != mcast_addr_))
+            {
+                local_segment.push_back(p.socket().get());
+                if (p.remote_uuid() < uuid())
+                {
+                    ++self_index_;
+                }
+            }
+        }
+        else
+        {
+            if (p.state() == Proto::S_OK)
+            {
+                Segment& remote_segment(segment_map_[p.remote_segment()]);
+                remote_segment.push_back(p.socket().get());
+            }
         }
     }
+    log_debug << self_string() << " self index: " << self_index_;
     log_debug << self_string() << " --- mcast tree end ---";
 }
 
@@ -945,7 +969,12 @@ namespace
     class CmpUuidCounts
     {
     public:
-        CmpUuidCounts(const std::set<gcomm::UUID>& uuids) : uuids_(uuids) { }
+        CmpUuidCounts(const std::set<gcomm::UUID>& uuids,
+                      gcomm::SegmentId preferred_segment)
+            :
+            uuids_(uuids),
+            preferred_segment_(preferred_segment)
+        { }
 
         size_t count(const gcomm::gmcast::Proto* p) const
         {
@@ -969,13 +998,16 @@ namespace
         bool operator()(const gcomm::gmcast::Proto* a,
                         const gcomm::gmcast::Proto* b) const
         {
-
-            return (count(a) < count(b));
-
+            size_t ac(count(a));
+            size_t bc(count(b));
+            // if counts are equal, prefer peer from the same segment
+            return (ac < bc ||
+                    (ac == bc && a->remote_segment() != preferred_segment_));
         }
 
     private:
         const std::set<gcomm::UUID>& uuids_;
+        gcomm::SegmentId preferred_segment_;
     };
 }
 
@@ -1005,8 +1037,19 @@ void gcomm::GMCast::check_liveness()
         }
         else if (p->state() == Proto::S_OK)
         {
-            // log_info << "live proto " << *p;
-            live_uuids.insert(p->remote_uuid());
+            if (p->tstamp() + peer_timeout_*2/3 < now)
+            {
+                p->send_keepalive();
+            }
+
+            if (p->state() == Proto::S_FAILED)
+            {
+                handle_failed(p);
+            }
+            else
+            {
+                live_uuids.insert(p->remote_uuid());
+            }
         }
         i = i_next;
     }
@@ -1061,10 +1104,10 @@ void gcomm::GMCast::check_liveness()
         {
             std::set<Proto*>::iterator maxel(
                 std::max_element(proto_set.begin(),
-                                 proto_set.end(), CmpUuidCounts(nonlive_uuids)));
+                                 proto_set.end(), CmpUuidCounts(nonlive_uuids, segment_)));
             Proto* p(*maxel);
             log_debug << "relay set maxel :" << *p << " count: "
-                      << CmpUuidCounts(nonlive_uuids).count(p);
+                      << CmpUuidCounts(nonlive_uuids, segment_).count(p);
 
             relay_set_.insert(p->socket().get());
             const LinkMap& lm(p->link_map());
@@ -1100,24 +1143,81 @@ gu::datetime::Date gcomm::GMCast::handle_timers()
 }
 
 
-void gcomm::GMCast::relay(const Message& msg, const Datagram& dg,
+void send(gcomm::Socket* s, gcomm::Datagram& dg)
+{
+    int err;
+    if ((err = s->send(dg)) != 0)
+    {
+        log_debug << "failed to send to " << s->remote_addr()
+                  << ": (" << err << ") " << strerror(err);
+    }
+}
+
+void gcomm::GMCast::relay(const Message& msg,
+                          const Datagram& dg,
                           const void* exclude_id)
 {
-    Message relay_msg(msg);
-    relay_msg.set_flags(relay_msg.flags() & ~Message::F_RELAY);
     Datagram relay_dg(dg);
     relay_dg.normalize();
-    gu_trace(push_header(relay_msg, relay_dg));
-    for (std::list<Socket*>::iterator i(mcast_tree_.begin());
-         i != mcast_tree_.end(); ++i)
+    Message relay_msg(msg);
+
+    // reset all relay flags from message to be relayed
+    relay_msg.set_flags(relay_msg.flags() &
+                        ~(Message::F_RELAY | Message::F_SEGMENT_RELAY));
+
+    // if F_RELAY is set in received message, relay to all peers except
+    // the originator
+    if (msg.flags() & Message::F_RELAY)
     {
-        int err;
-        if ((*i)->id() != exclude_id &&
-            (err = (*i)->send(relay_dg)) != 0)
+        gu_trace(push_header(relay_msg, relay_dg));
+        for (SegmentMap::iterator i(segment_map_.begin());
+             i != segment_map_.end(); ++i)
         {
-            log_debug << "transport: " << ::strerror(err);
+            Segment& segment(i->second);
+            for (Segment::iterator j(segment.begin()); j != segment.end(); ++j)
+            {
+                if ((*j)->id() != exclude_id)
+                {
+                    send(*j, relay_dg);
+                }
+            }
         }
     }
+    else if (msg.flags() & Message::F_SEGMENT_RELAY)
+    {
+        if (relay_set_.empty() == false)
+        {
+            // send message to all nodes in relay set to reach
+            // nodes in local segment that are not directly reachable
+            relay_msg.set_flags(relay_msg.flags() | Message::F_RELAY);
+            gu_trace(push_header(relay_msg, relay_dg));
+            for (std::set<Socket*>::iterator ri(relay_set_.begin());
+                 ri != relay_set_.end(); ++ri)
+            {
+                send(*ri, relay_dg);
+            }
+            gu_trace(pop_header(relay_msg, relay_dg));
+            relay_msg.set_flags(relay_msg.flags() & ~Message::F_RELAY);
+        }
+
+        if (msg.segment_id() == segment_)
+        {
+            log_warn << "message with F_SEGMENT_RELAY from own segment, "
+                     << "source " << msg.source_uuid();
+        }
+
+        // Relay to local segment
+        gu_trace(push_header(relay_msg, relay_dg));
+        Segment& segment(segment_map_[segment_]);
+        for (Segment::iterator i(segment.begin()); i != segment.end(); ++i)
+        {
+            send(*i, relay_dg);
+        }
+    }
+    else
+    {
+        log_warn << "GMCast::relay() called without relay flags set";
+    }
 }
 
 void gcomm::GMCast::handle_up(const void*        id,
@@ -1202,22 +1302,24 @@ void gcomm::GMCast::handle_up(const void*        id,
 
             if (msg.type() >= Message::T_USER_BASE)
             {
-                if (msg.flags() & Message::F_RELAY)
+                if (msg.flags() &
+                    (Message::F_RELAY | Message::F_SEGMENT_RELAY))
                 {
                     relay(msg,
                           Datagram(dg, dg.offset() + msg.serial_size()),
                           id);
                 }
+                p->set_tstamp(gu::datetime::Date::now());
                 send_up(Datagram(dg, dg.offset() + msg.serial_size()),
                         ProtoUpMeta(msg.source_uuid()));
-                p->set_tstamp(gu::datetime::Date::now());
+                return;
             }
             else
             {
                 try
                 {
-                    gu_trace(p->handle_message(msg));
                     p->set_tstamp(gu::datetime::Date::now());
+                    gu_trace(p->handle_message(msg));
                 }
                 catch (gu::Exception& e)
                 {
@@ -1271,37 +1373,63 @@ void gcomm::GMCast::handle_up(const void*        id,
     }
 }
 
+
 int gcomm::GMCast::handle_down(Datagram& dg, const ProtoDownMeta& dm)
 {
-    Message msg(version_, Message::T_USER_BASE, uuid(), 1);
-    gu_trace(push_header(msg, dg));
+    Message msg(version_, Message::T_USER_BASE, uuid(), 1, segment_);
 
-    for (std::list<Socket*>::iterator i(mcast_tree_.begin());
-         i != mcast_tree_.end(); ++i)
+    // handle relay set first, skip these peers below
+    if (relay_set_.empty() == false)
     {
-        bool relay(false);
-        if (relay_set_.empty()  == false &&
-            relay_set_.find(*i) != relay_set_.end())
+        msg.set_flags(msg.flags() | Message::F_RELAY);
+        gu_trace(push_header(msg, dg));
+        for (std::set<Socket*>::iterator ri(relay_set_.begin());
+             ri != relay_set_.end(); ++ri)
         {
-            relay = true;
-            gu_trace(pop_header(msg, dg));
-            msg.set_flags(msg.flags() | Message::F_RELAY);
-            gu_trace(push_header(msg, dg));
-        }
-        int err;
-        if ((err = (*i)->send(dg)) != 0)
-        {
-            log_debug << "transport: " << ::strerror(err);
-        }
-        if (relay == true)
-        {
-            gu_trace(pop_header(msg, dg));
-            msg.set_flags(msg.flags() & ~Message::F_RELAY);
-            gu_trace(push_header(msg, dg));
+            send(*ri, dg);
         }
+        gu_trace(pop_header(msg, dg));
+        msg.set_flags(msg.flags() & ~Message::F_RELAY);
     }
 
-    gu_trace(pop_header(msg, dg));
+
+
+    for (SegmentMap::iterator si(segment_map_.begin());
+         si != segment_map_.end(); ++si)
+    {
+        uint8_t segment_id(si->first);
+        Segment& segment(si->second);
+
+        if (segment_id != segment_)
+        {
+            size_t target_idx((self_index_ + segment_id) % segment.size());
+            msg.set_flags(msg.flags() | Message::F_SEGMENT_RELAY);
+            // skip peers that are in relay set
+            if (relay_set_.empty() == true ||
+                relay_set_.find(segment[target_idx]) == relay_set_.end())
+            {
+                gu_trace(push_header(msg, dg));
+                send(segment[target_idx], dg);
+                gu_trace(pop_header(msg, dg));
+            }
+        }
+        else
+        {
+            msg.set_flags(msg.flags() & ~Message::F_SEGMENT_RELAY);
+            gu_trace(push_header(msg, dg));
+            for (Segment::iterator i(segment.begin());
+                 i != segment.end(); ++i)
+            {
+                // skip peers that are in relay set
+                if (relay_set_.empty() == true ||
+                    relay_set_.find(*i) == relay_set_.end())
+                {
+                    send(*i, dg);
+                }
+            }
+            gu_trace(pop_header(msg, dg));
+        }
+    }
 
     return 0;
 }
@@ -1496,7 +1624,7 @@ bool gcomm::GMCast::set_param(const std::string& key, const std::string& val)
                 delete rp;
                 proto_map_->erase(pi);
             }
-            mcast_tree_.clear();
+            segment_map_.clear();
         }
         return true;
     }
@@ -1506,7 +1634,8 @@ bool gcomm::GMCast::set_param(const std::string& key, const std::string& val)
              key == Conf::GMCastMCastPort ||
              key == Conf::GMCastMCastTTL ||
              key == Conf::GMCastTimeWait ||
-             key == Conf::GMCastPeerTimeout)
+             key == Conf::GMCastPeerTimeout ||
+             key == Conf::GMCastSegment)
     {
         gu_throw_error(EPERM) << "can't change value for '"
                               << key << "' during runtime";
