@@ -1,7 +1,7 @@
 /*
- * Copyright (C) 2008-2013 Codership Oy <info@codership.com>
+ * Copyright (C) 2008-2014 Codership Oy <info@codership.com>
  *
- * $Id: gcs.c 3336 2013-10-28 07:41:56Z teemu $
+ * $Id: gcs.c 3455 2014-02-01 18:11:09Z alex $
  */
 
 /*
@@ -168,9 +168,10 @@ struct gcs_recv_act
 
 struct gcs_repl_act
 {
-    struct gcs_action* action;
-    gu_mutex_t         wait_mutex;
-    gu_cond_t          wait_cond;
+    const struct gu_buf* act_in;
+    struct gcs_action*   action;
+    gu_mutex_t           wait_mutex;
+    gu_cond_t            wait_cond;
 };
 
 /*! Releases resources associated with parameters */
@@ -1031,7 +1032,7 @@ _handle_timeout (gcs_conn_t* conn)
     bool ret;
     long long now = gu_time_calendar();
 
-    /* TODO: now the only point for timeout is flow control (#412), 
+    /* TODO: now the only point for timeout is flow control (#412),
      *       later we might need to handle more timers. */
     if (conn->timeout <= now) {
         ret = ((GCS_CONN_JOINER != conn->state) ||
@@ -1156,12 +1157,12 @@ static void *gcs_recv_thread (void *arg)
         if (gu_likely (rcvd.act.type != GCS_ACT_TORDERED ||
                        (rcvd.id > 0 && (conn->global_seqno = rcvd.id)))) {
             /* successful delivery - increment local order */
-            this_act_id = conn->local_act_id++;
+            this_act_id = gu_sync_fetch_and_add(&conn->local_act_id, 1);
         }
 
-        if (NULL != rcvd.repl_buf                                       &&
+        if (NULL != rcvd.local                                          &&
             (repl_act_ptr = gcs_fifo_lite_get_head (conn->repl_q))      &&
-            (gu_likely ((*repl_act_ptr)->action->buf == rcvd.repl_buf)  ||
+            (gu_likely ((*repl_act_ptr)->act_in == rcvd.local)  ||
              /* at this point repl_q is locked and we need to unlock it and
               * return false to fall to the 'else' branch; unlikely case */
              (gcs_fifo_lite_release (conn->repl_q), false)))
@@ -1417,11 +1418,11 @@ long gcs_destroy (gcs_conn_t *conn)
 }
 
 /* Puts action in the send queue and returns */
-long gcs_send (gcs_conn_t*    const conn,
-               const void*    const act_buf,
-               size_t         const act_size,
-               gcs_act_type_t const act_type,
-               bool           const scheduled)
+long gcs_sendv (gcs_conn_t*          const conn,
+                const struct gu_buf* const act_bufs,
+                size_t               const act_size,
+                gcs_act_type_t       const act_type,
+                bool                 const scheduled)
 {
     if (gu_unlikely(act_size > GCS_MAX_ACT_SIZE)) return -EMSGSIZE;
 
@@ -1436,7 +1437,7 @@ long gcs_send (gcs_conn_t*    const conn,
     if (!(ret = gcs_sm_enter (conn->sm, &tmp_cond, scheduled)))
     {
         while ((GCS_CONN_OPEN >= conn->state) &&
-               (ret = gcs_core_send (conn->core, act_buf,
+               (ret = gcs_core_send (conn->core, act_bufs,
                                      act_size, act_type)) == -ERESTART);
         gcs_sm_leave (conn->sm);
         gu_cond_destroy (&tmp_cond);
@@ -1461,23 +1462,23 @@ gcs_seqno_t gcs_caused(gcs_conn_t* conn)
 }
 
 /* Puts action in the send queue and returns after it is replicated */
-long gcs_repl (gcs_conn_t*        conn,      //!<in
-               struct gcs_action* act,       //!<inout
-               bool               scheduled) //!<in
+long gcs_replv (gcs_conn_t*          const conn,      //!<in
+                const struct gu_buf* const act_in,    //!<in
+                struct gcs_action*   const act,       //!<inout
+                bool                 const scheduled) //!<in
 {
     if (gu_unlikely((size_t)act->size > GCS_MAX_ACT_SIZE)) return -EMSGSIZE;
 
     long ret;
 
     assert (act);
-    assert (act->buf);
     assert (act->size > 0);
 
     act->seqno_l = GCS_SEQNO_ILL;
     act->seqno_g = GCS_SEQNO_ILL;
 
     /* This is good - we don't have to do a copy because we wait */
-    struct gcs_repl_act repl_act = { .action = act };
+    struct gcs_repl_act repl_act = { .act_in = act_in, .action = act };
 
     gu_mutex_init (&repl_act.wait_mutex, NULL);
     gu_cond_init  (&repl_act.wait_cond,  NULL);
@@ -1512,7 +1513,7 @@ long gcs_repl (gcs_conn_t*        conn,      //!<in
                 gcs_fifo_lite_push_tail (conn->repl_q);
 
                 // Keep on trying until something else comes out
-                while ((ret = gcs_core_send (conn->core, act->buf, act->size,
+                while ((ret = gcs_core_send (conn->core, act_in, act->size,
                                              act->type)) == -ERESTART) {}
 
                 if (ret < 0) {
@@ -1562,9 +1563,12 @@ long gcs_repl (gcs_conn_t*        conn,      //!<in
                     }
 
                     if (orig_buf != act->buf) // action was allocated in gcache
+                    {
+                        gu_debug("Freeing gcache buffer %p after receiving %d",
+                                 act->buf, ret);
                         gcs_gcache_free (conn->gcache, act->buf);
-
-                    act->buf = NULL;
+                        act->buf = orig_buf;
+                    }
                 }
             }
         }
@@ -1610,15 +1614,15 @@ long gcs_request_state_transfer (gcs_conn_t  *conn,
 
         ret = gcs_repl(conn, &action, false);
 
-        assert (action.buf != rst);
-
         gu_free (rst);
 
         *local = action.seqno_l;
 
         if (ret > 0) {
+            assert (action.buf != rst);
 #ifndef GCS_FOR_GARB
             assert (action.buf != NULL);
+            gcs_gcache_free (conn->gcache, action.buf);
 #else
             assert (action.buf == NULL);
 #endif
@@ -1626,13 +1630,12 @@ long gcs_request_state_transfer (gcs_conn_t  *conn,
             assert (action.seqno_g >= 0);
             assert (action.seqno_l >  0);
 
-            gcs_gcache_free (conn->gcache, action.buf);
             // on joiner global seqno stores donor index
             // on donor global seqno stores global seqno
             ret = action.seqno_g;
         }
         else {
-            assert (action.buf == NULL);
+            assert (/*action.buf == NULL ||*/ action.buf == rst);
         }
     }
 
@@ -1811,22 +1814,37 @@ gcs_join (gcs_conn_t* conn, gcs_seqno_t seqno)
     return _join (conn, seqno);
 }
 
+gcs_seqno_t gcs_local_sequence(gcs_conn_t* conn)
+{
+    return gu_sync_fetch_and_add(&conn->local_act_id, 1);
+}
+
 void
 gcs_get_stats (gcs_conn_t* conn, struct gcs_stats* stats)
 {
-    gu_fifo_stats (conn->recv_q,
-                   &stats->recv_q_len,
-                   &stats->recv_q_len_avg);
+    gu_fifo_stats_get (conn->recv_q,
+                       &stats->recv_q_len,
+                       &stats->recv_q_len_avg);
 
     stats->recv_q_size = conn->recv_q_size;
 
-    gcs_sm_stats  (conn->sm,
-                   &stats->send_q_len,
-                   &stats->send_q_len_avg,
-                   &stats->fc_paused);
+    gcs_sm_stats_get (conn->sm,
+                      &stats->send_q_len,
+                      &stats->send_q_len_avg,
+                      &stats->fc_paused_ns,
+                      &stats->fc_paused_avg);
 
-    stats->fc_sent     = conn->stats_fc_sent;     conn->stats_fc_sent     = 0;
-    stats->fc_received = conn->stats_fc_received; conn->stats_fc_received = 0;
+    stats->fc_sent     = conn->stats_fc_sent;
+    stats->fc_received = conn->stats_fc_received;
+}
+
+void
+gcs_flush_stats(gcs_conn_t* conn)
+{
+    gu_fifo_stats_flush(conn->recv_q);
+    gcs_sm_stats_flush (conn->sm);
+    conn->stats_fc_sent     = 0;
+    conn->stats_fc_received = 0;
 }
 
 static long
@@ -2025,6 +2043,12 @@ _set_max_throttle (gcs_conn_t* conn, const char* value)
     }
 }
 
+void gcs_register_params (gu_config_t* const conf)
+{
+    gcs_params_register (conf);
+    gcs_core_register   (conf);
+}
+
 long gcs_param_set  (gcs_conn_t* conn, const char* key, const char *value)
 {
     if (!strcmp (key, GCS_PARAMS_FC_LIMIT)) {
