@@ -1,5 +1,5 @@
 /*
- * Copyright (C) 2009 Codership Oy <info@codership.com>
+ * Copyright (C) 2009-2014 Codership Oy <info@codership.com>
  */
 
 /*!
@@ -28,6 +28,7 @@ extern "C"
 #include "gu_backtrace.hpp"
 #include "gcomm/transport.hpp"
 #include "gcomm/util.hpp"
+#include "gcomm/conf.hpp"
 
 #ifdef PROFILE_GCS_GCOMM
 #define GCOMM_PROFILE 1
@@ -207,7 +208,9 @@ public:
         return 0;
     }
 
-    void connect(const string& channel)
+    void connect(bool) { }
+
+    void connect(const string& channel, bool const bootstrap)
     {
         if (tp_ != 0)
         {
@@ -218,26 +221,35 @@ public:
         tp_ = Transport::create(*net_, uri_);
         gcomm::connect(tp_, this);
 
-        string peer;
-        URI::AuthorityList::const_iterator i, i_next;
-        for (i = uri_.get_authority_list().begin();
-             i != uri_.get_authority_list().end(); ++i)
+        if (bootstrap)
         {
-            i_next = i;
-            ++i_next;
-            string host;
-            string port;
-            try { host = i->host(); } catch (NotSet&) { }
-            try { port = i->port(); } catch (NotSet&) { }
-            peer += host != "" ? host + ":" + port : "";
-            if (i_next != uri_.get_authority_list().end())
-            {
-                peer += ",";
-            }
+            log_info << "gcomm: bootstrapping new group '" << channel << '\'';
         }
-        log_info << "gcomm: connecting to group '" << channel
-                 << "', peer '" << peer << "'";
-        tp_->connect();
+        else
+        {
+            string peer;
+            URI::AuthorityList::const_iterator i, i_next;
+            for (i = uri_.get_authority_list().begin();
+                 i != uri_.get_authority_list().end(); ++i)
+            {
+                i_next = i;
+                ++i_next;
+                string host;
+                string port;
+                try { host = i->host(); } catch (NotSet&) { }
+                try { port = i->port(); } catch (NotSet&) { }
+                peer += host != "" ? host + ":" + port : "";
+                if (i_next != uri_.get_authority_list().end())
+                {
+                    peer += ",";
+                }
+            }
+            log_info << "gcomm: connecting to group '" << channel
+                     << "', peer '" << peer << "'";
+        }
+
+        tp_->connect(bootstrap);
+
         uuid_ = tp_->uuid();
 
         int err;
@@ -249,7 +261,7 @@ public:
         log_info << "gcomm: connected";
     }
 
-    void close()
+    void close(bool force = false)
     {
         if (tp_ == 0)
         {
@@ -303,9 +315,11 @@ public:
         }
         return tp_->mtu();
     }
+
     Protonet&   get_pnet()                { return *net_; }
     gu::Config& get_conf()                { return conf_; }
     int         get_error() const         { return error_; }
+
     class Ref
     {
     public:
@@ -441,7 +455,7 @@ void GCommConn::run()
         }
         catch (gu::Exception& e)
         {
-            log_error << "exception from gcomm, backend must be restarted:"
+            log_error << "exception from gcomm, backend must be restarted: "
                       << e.what();
             // Commented out due to Backtrace() not producing proper
             // backtraces.
@@ -545,7 +559,8 @@ static void fill_cmp_msg(const View& view, const UUID& my_uuid,
 
 //        (void)snprintf(cm->memb[n].id, GCS_COMP_MEMB_ID_MAX_LEN, "%s",
 //                       uuid._str().c_str());
-        long ret = gcs_comp_msg_add (cm, uuid._str().c_str());
+        long ret = gcs_comp_msg_add (cm, uuid._str().c_str(),
+                                     i->second.segment());
         if (ret < 0) {
             gu_throw_error(-ret) << "Failed to add member '" << uuid
                                  << "' to component message.";
@@ -687,7 +702,7 @@ static GCS_BACKEND_OPEN_FN(gcomm_open)
 
     try
     {
-        conn.connect(channel);
+        conn.connect(channel, bootstrap);
     }
     catch (Exception& e)
     {
@@ -820,7 +835,10 @@ GCS_BACKEND_PARAM_GET_FN(gcomm_param_get)
 }
 
 
-
+GCS_BACKEND_REGISTER_FN(gcs_gcomm_register)
+{
+    gcomm::Conf::register_params(*reinterpret_cast<gu::Config*>(cnf));
+}
 
 
 GCS_BACKEND_CREATE_FN(gcs_gcomm_create)
@@ -835,7 +853,7 @@ GCS_BACKEND_CREATE_FN(gcs_gcomm_create)
 
     try
     {
-        gu::URI uri(std::string("pc://") + socket);
+        gu::URI uri(std::string("pc://") + addr);
         gu::Config& conf(*reinterpret_cast<gu::Config*>(cnf));
         conn = new GCommConn(uri, conf);
     }
