@@ -1,7 +1,7 @@
 /*
- * Copyright (C) 2008 Codership Oy <info@codership.com>
+ * Copyright (C) 2008-2012 Codership Oy <info@codership.com>
  *
- * $Id: gcs_test.c 2745 2012-03-17 00:00:23Z alex $
+ * $Id: gcs_test.c 3349 2013-11-02 11:36:15Z alex $
  */
 /***********************************************************/
 /*  This program imitates 3rd party application and        */
@@ -204,6 +204,7 @@ test_log_open (gcs_test_log_t **log, const char *name)
     if (!l) return errno;
 
     snprintf (real_name, 1024, "%s.%lld", name, (long long)getpid());
+    // cppcheck-suppress memleak
     if (!(l->file = fopen (real_name, "w"))) return errno;
     pthread_mutex_init (&l->lock, NULL);
     *log = l;
@@ -707,7 +708,9 @@ int main (int argc, char *argv[])
     if (!(gcache = gcache_create (gconf, ""))) goto out;
     if (!(gcs = gcs_create (gconf, gcache, NULL, NULL, 0, 0))) goto out;
     puts ("debug"); fflush(stdout);
-    if ((err  = gcs_open   (gcs, channel, conf.backend))) goto out;
+    /* the following hack won't work if there is 0.0.0.0 in URL options */
+    bool bstrap = (NULL != strstr(conf.backend, "0.0.0.0"));
+    if ((err  = gcs_open   (gcs, channel, conf.backend, bstrap))) goto out;
     printf ("Connected\n");
 
     msg_len = 1300;
