@@ -9,7 +9,7 @@ fi
 use_mysql_5.1_sources()
 {
     MYSQL_MAJOR="5.1"
-    export MYSQL_5_1=$MYSQL_MAJOR # for DEB build
+    export MYSQL_MAJOR # for DEB build
     export MYSQL_MAJOR_VER="5"
     export MYSQL_MINOR_VER="1"
     MYSQL_VER=`grep AC_INIT $MYSQL_SRC/configure.in | awk -F '[' '{ print $3 }' | awk -F ']' '{ print $1 }'`
@@ -26,7 +26,7 @@ use_mysql_5.5_sources()
     export MYSQL_MINOR_VER=`grep MYSQL_VERSION_MINOR $MYSQL_SRC/VERSION | cut -d = -f 2`
     export MYSQL_PATCH_VER=`grep MYSQL_VERSION_PATCH $MYSQL_SRC/VERSION | cut -d = -f 2`
     MYSQL_MAJOR=$MYSQL_MAJOR_VER.$MYSQL_MINOR_VER
-    export MYSQL_5_5=$MYSQL_MAJOR # for DEB build
+    export MYSQL_MAJOR # for DEB build
     MYSQL_VER=$MYSQL_MAJOR.$MYSQL_PATCH_VER
 }
 
@@ -45,7 +45,7 @@ fi
 uname -m | grep -q i686 && CPU=pentium || CPU=amd64 # this works for x86 Solaris too
 BOOTSTRAP=no
 DEBUG=no
-DEBUG_LEVEL=1
+DEBUG_LEVEL=0
 GALERA_DEBUG=no
 NO_STRIP=no
 RELEASE=""
@@ -61,6 +61,9 @@ SCRATCH=no
 SCONS="yes"
 JOBS=1
 GCOMM_IMPL=${GCOMM_IMPL:-"galeracomm"}
+TARGET=""
+MYSQLD_BINARY="mysqld"
+SYNC_BEFORE_PACK=${SYNC_BEFORE_PACK:-""}
 
 OS=$(uname)
 case "$OS" in
@@ -292,7 +295,7 @@ then
 else
     [ "$DEBUG" == "yes" ] \
     && MYSQL_BUILD_DIR="$MYSQL_SRC/build_debug" \
-    || MYSQL_BUILD_DIR="$MYSQL_SRC/build_install"
+    || MYSQL_BUILD_DIR="$MYSQL_SRC/build_release"
 fi
 
 
@@ -311,9 +314,9 @@ then
         debug_opt="-d"
     fi
     scripts/build.sh $debug_opt # options are passed via environment variables
-    GALERA_REV=$(bzr revno 2>/dev/null)     || \
-    GALERA_REV=$(svnversion | sed s/\:/,/g) || \
-    GALERA_REV=$(echo "xxxx")
+    GALERA_REV=$(bzr revno --tree -q) || \
+    GALERA_REV=$(svn info >&/dev/null && svnversion | sed s/\:/,/g) || \
+    GALERA_REV=$(echo "XXXX")
 fi
 
 ######################################
@@ -323,14 +326,14 @@ fi
 ######################################
 # Obtain MySQL version and revision number
 cd $MYSQL_SRC
-WSREP_REV=$(bzr revno) || \
+WSREP_REV=$(bzr revno --tree -q) || \
 WSREP_REV="XXXX"
 # this does not work on an unconfigured source MYSQL_VER=$(grep '#define VERSION' $MYSQL_SRC/include/config.h | sed s/\"//g | cut -d ' ' -f 3 | cut -d '-' -f 1-2)
 
 if [ "$PACKAGE" == "yes" ] || [ "$BIN_DIST" == "yes" ]
 then
     # fetch and patch pristine sources
-    cd /tmp
+    cd ${TMPDIR:-/tmp}
     mysql_tag=mysql-$MYSQL_VER
     if [ "$SKIP_BUILD" == "no" ] || [ ! -d $mysql_tag ]
     then
@@ -349,12 +352,13 @@ then
         tar -xzf $mysql_orig_tar_gz
         cd $mysql_tag/
         patch -p1 -f < $patch_file >/dev/null || :
-        chmod a+x ./BUILD/*wsrep
+        # chmod a+x ./BUILD/*wsrep
         CONFIGURE="yes"
     else
         cd $mysql_tag/
     fi
     MYSQL_SRC=$(pwd -P)
+    MYSQL_BUILD_DIR=$MYSQL_SRC
     if [ "$CONFIGURE" == "yes" ]
     then
         echo "Regenerating config files"
@@ -404,15 +408,17 @@ then
             [ $DEBIAN -ne 0 ] && \
                 MYSQL_SOCKET_PATH="/var/run/mysqld/mysqld.sock" || \
                 MYSQL_SOCKET_PATH="/var/lib/mysql/mysql.sock"
+
             CMAKE_LAYOUT_OPTIONS=( \
-                -DCMAKE_INSTALL_PREFIX="/usr" \
                 -DINSTALL_LAYOUT=RPM \
+                -DCMAKE_INSTALL_PREFIX="/usr" \
                 -DINSTALL_SBINDIR="/usr/sbin" \
                 -DMYSQL_DATADIR="/var/lib/mysql" \
                 -DMYSQL_UNIX_ADDR=$MYSQL_SOCKET_PATH \
                 -DCMAKE_OSX_ARCHITECTURES=$(uname -m) \
             )
         fi
+
         [ -n "$EXTRA_SYSROOT" ] && \
             CMAKE_LAYOUT_OPTIONS+=( \
                 -DCMAKE_PREFIX_PATH="$EXTRA_SYSROOT" \
@@ -432,26 +438,37 @@ then
             BUILD/compile-${CPU}${BUILD_OPT}-wsrep > /dev/null
         else # CMake build
             [ "$DEBUG" = "yes" ] \
-            && BUILD_OPT="-DCMAKE_BUILD_TYPE=Debug" \
-            || BUILD_OPT="-DBUILD_CONFIG=mysql_release"
-            [ "$BOOTSTRAP" = "yes" ] && rm -rf $MYSQL_BUILD_DIR
-            [ -d "$MYSQL_BUILD_DIR" ] || mkdir -p $MYSQL_BUILD_DIR
+            && BUILD_OPT="-DCMAKE_BUILD_TYPE=Debug -DDEBUG_EXTNAME=OFF" \
+            || BUILD_OPT="-DCMAKE_BUILD_TYPE=RelWithDebInfo" # like in RPM spec
+
+            [ "$MYSQL_MAJOR_VER$MYSQL_MINOR_VER" -ge "56" ] \
+            && MEMCACHED_OPT="-DWITH_LIBEVENT=yes -DWITH_INNODB_MEMCACHED=ON" \
+            || MEMCACHED_OPT=""
+
+            if [ "$MYSQL_BUILD_DIR" != "$MYSQL_SRC" ]
+            then
+               [ "$BOOTSTRAP" = "yes" ] && rm -rf $MYSQL_BUILD_DIR
+               [ -d "$MYSQL_BUILD_DIR" ] || mkdir -p $MYSQL_BUILD_DIR
+            fi
+
             pushd $MYSQL_BUILD_DIR
-            cmake $BUILD_OPT \
+            cmake \
                   ${CC:+-DCMAKE_C_COMPILER="$CC"} \
                   ${CXX:+-DCMAKE_CXX_COMPILER="$CXX"} \
-                  -DWITH_WSREP=1 \
+                  -DBUILD_CONFIG=mysql_release \
                   "${CMAKE_LAYOUT_OPTIONS[@]}" \
+                  $BUILD_OPT \
+                  -DWITH_WSREP=1 \
                   -DWITH_EXTRA_CHARSETS=all \
-                  -DWITH_READLINE=yes \
-                  -DWITH_SSL=system \
+                  -DWITH_SSL=yes \
                   -DWITH_ZLIB=system \
+                  $MEMCACHED_OPT \
                   $MYSQL_SRC \
-            && make -S && popd || exit 1
+            && make -j $JOBS -S && popd || exit 1
         fi
     else  # just recompile and relink with old configuration
         [ $MYSQL_MAJOR != "5.1" ] && pushd $MYSQL_BUILD_DIR
-        make -S > /dev/null
+        make -j $JOBS -S > /dev/null
         [ $MYSQL_MAJOR != "5.1" ] && popd
     fi
 fi # SKIP_BUILD
@@ -563,10 +580,10 @@ install_mysql_5.5_demo()
     popd
     unset DESTDIR
     pushd $MYSQL_DIST_DIR
-    mv usr/local/* ./ && rmdir usr/local # FreeBSD
-    [ -d libexec -a ! -a sbin ] && mv libexec sbin # FreeBSD
+    [ -d usr/local ] && ( mv usr/local/* ./ && rmdir usr/local ) # FreeBSD
+    [ -d libexec -a ! -d sbin ] && mv libexec sbin # FreeBSD
     mv usr/* ./ && rmdir usr
-    [ -d lib64 -a ! -a lib ] && mv lib64 lib
+    [ -d lib64 -a ! -d lib ] && mv lib64 lib
     popd
 }
 
@@ -659,9 +676,12 @@ if [ "$TAR" == "yes" ] || [ "$BIN_DIST" == "yes" ]; then
     mv $DIST_DIR $RELEASE_NAME
 
     # Hack to avoid 'file changed as we read it'-error
-    sync
-    sleep 1
-
+    if test -n "$SYNC_BEFORE_PACK"
+    then
+        echo "syncing disks"
+        sync
+        sleep 1
+    fi
     # Pack the release
     tar -czf $RELEASE_NAME.tgz $RELEASE_NAME
 
@@ -675,7 +695,15 @@ fi
 
 get_arch()
 {
-    if [ "$OS" == "Darwin" ]; then
+    if ! [ -z "$TARGET" ]
+    then
+        if [ "$TARGET" == "i686" ]
+        then
+            echo "i386"
+        else
+            echo "amd64"
+        fi
+    elif [ "$OS" == "Darwin" ]; then
         if file $MYSQL_SRC/sql/mysqld | grep "i386" >/dev/null 2>&1
         then
             echo "i386"
@@ -707,7 +735,7 @@ build_linux_packages()
     local STRIP_OPT=""
     [ "$NO_STRIP" == "yes" ] && STRIP_OPT="-g"
 
-    export MYSQL_VER MYSQL_SRC GALERA_SRC RELEASE_NAME
+    export MYSQL_VER MYSQL_SRC GALERA_SRC RELEASE_NAME MYSQLD_BINARY
     export WSREP_VER=${RELEASE:-"$WSREP_REV"}
 
     echo $MYSQL_SRC $MYSQL_VER $ARCH
@@ -736,8 +764,11 @@ build_freebsd_packages()
     # Create build directory structure
     DIST_DIR=$BUILD_ROOT/dist/mysql
     MYSQL_DIST_DIR=$DIST_DIR/usr/local
-    MYSQL_DIST_CNF=$MYSQL_DIST_DIR/etc/my.cnf
+    MYSQL_DIST_CNF=$MYSQL_DIST_DIR/share/mysql/my_wsrep.cnf
     MYSQL_BINS=$MYSQL_DIST_DIR/bin
+    MYSQL_CLIENT_LICENSE_DIR=$MYSQL_DIST_DIR/share/licenses/mysql-client-${MYSQL_VER}_wsrep_${RELEASE}
+    MYSQL_SERVER_LICENSE_DIR=$MYSQL_DIST_DIR/share/licenses/mysql-server-${MYSQL_VER}_wsrep_${RELEASE}
+    MYSQL_SERVER_DOC_DIR=$MYSQL_DIST_DIR/share/doc/mysql${MYSQL_MAJOR_VER}${MYSQL_MINOR_VER}-server_wsrep
 
     cd $BUILD_ROOT
     rm -rf $BUILD_ROOT/dist
@@ -749,12 +780,20 @@ build_freebsd_packages()
     cat $MYSQL_BUILD_DIR/support-files/wsrep.cnf | \
         sed 's/root:$/root:rootpass/' >> $MYSQL_DIST_CNF
     pushd $MYSQL_BINS; ln -s wsrep_sst_rsync wsrep_sst_rsync_wan; popd
-    install -m 755 -d $MYSQL_DIST_DIR/share/doc/mysql
-    install -m 644 LICENSE.mysql $MYSQL_DIST_DIR/share/doc/mysql
 
-    install -m 644 LICENSE     $DIST_DIR/usr/local/share/doc/mysql/LICENSE.wsrep
-    install -m 644 README      $DIST_DIR/usr/local/share/doc/mysql/README.wsrep
-    install -m 644 QUICK_START $DIST_DIR/usr/local/share/doc/mysql/QUICK_START.wsrep
+    install -m 755 -d "$MYSQL_CLIENT_LICENSE_DIR"
+    install -m 644 ../../LICENSE "$MYSQL_CLIENT_LICENSE_DIR/GPLv3"
+    install -m 644 freebsd/LICENSE "$MYSQL_CLIENT_LICENSE_DIR"
+    install -m 644 freebsd/catalog.mk "$MYSQL_CLIENT_LICENSE_DIR"
+
+    install -m 755 -d "$MYSQL_SERVER_LICENSE_DIR"
+    install -m 644 ../../LICENSE "$MYSQL_SERVER_LICENSE_DIR/GPLv3"
+    install -m 644 freebsd/LICENSE "$MYSQL_SERVER_LICENSE_DIR"
+    install -m 644 freebsd/catalog.mk "$MYSQL_SERVER_LICENSE_DIR"
+
+    install -m 755 -d "$MYSQL_SERVER_DOC_DIR"
+    install -m 644 README      "$MYSQL_SERVER_DOC_DIR"
+    install -m 644 QUICK_START "$MYSQL_SERVER_DOC_DIR"
 
     # Strip binaries if not instructed otherwise
     if test "$NO_STRIP" != "yes"
